'''
The external_functions module creates the link between a function mentioned in the language measures and the python programme.
A function mentioned in a pre or core language measure takes as input a syntactic structure and
yields a list of nodes  as output.

A function mentioned in a post or form query takes as input an AllResults object and a
syntactic structure and yields  a result of any type as output.

If the function is defined in some other module, it must be imported here.
The name of the function must be added to one of the variables
*thetarspfunctions*, *thestapfunctions*, or *theastafunctions*, depending on the method that it belongs to.

'''

import re
from typing import Callable, Dict

from sastadev.allresults import QueryFunction
from sastadev.asta_queries import asta_bijzin, asta_delpv, asta_lex, asta_noun
from sastadev.astaforms import astaform
from sastadev.ASTApostfunctions import (KMcount, countwordsandcutoff,
                                        finietheidsindex, getalllemmas,
                                        getlexlemmas, getnounlemmas,
                                        neologisme, phonpar, sempar,
                                        wordcountperutt)
from sastadev.compounds import getcompounds
from sastadev.dedup import correct, mlux, onvolledig, samplesize
from sastadev.imperatives import wond4, wond5plus, wondx, wx, wxy, wxyz, wxyz5
from sastadev.methods import allok
from sastadev.queryfunctions import VzN, vobij, voslashbij, xneg_neg, xneg_x
from sastadev.stapforms import makestapform
from sastadev.STAPpostfunctions import GL5LVU, GLVU, BB_totaal
from sastadev.Sziplus import sziplus6, vr5plus
from sastadev.tarspform import mktarspform
from sastadev.TARSPpostfunctions import (gofase, gtotaal, pf, pf2, pf3, pf4,
                                         pf5, pf6, pf7, vutotaal)
from sastadev.TARSPscreening import tarsp_screening
from sastadev.xenx import xenx

normalfunctionpattern = r'<function\s+(\w+)\b'
builtinfunctionpattern = r'<built-in function\s+(\w+)\b'


# normalfunctionprefix = "<function "
# lnormalfunctionprefix = len(normalfunctionprefix)
# builtinfunctionprefix = "<built-in function "
# lbuiltinfunctionprefix = len(builtinfunctionprefix)

def getfname(f: Callable) -> str:
    return f.__name__


def oldgetfname(f: Callable) -> str:
    fstr = str(f)
    m = re.match(normalfunctionpattern, fstr)
    if m is not None:
        result = m.group(1)
    else:
        m = re.match(builtinfunctionpattern, fstr)
        if m is not None:
            result = m.group(1)
        else:
            result = ''
    return result


# Initialisation
thetarspfunctions = [getcompounds, sziplus6, xenx, vr5plus, wx, wxy, wxyz, wxyz5, wondx, wond4, wond5plus,
                     tarsp_screening, vutotaal, gofase, gtotaal, pf2, pf3, pf4, pf5, pf6, pf7, pf, xneg_x, xneg_neg,
                     mktarspform, VzN, vobij, voslashbij]

thestapfunctions = [BB_totaal, GLVU, GL5LVU, makestapform]

theastafunctions = [samplesize, mlux, neologisme, onvolledig, correct, wordcountperutt, countwordsandcutoff,
                    astaform, KMcount, finietheidsindex, getnounlemmas, getlexlemmas, getalllemmas, asta_noun,
                    asta_bijzin, asta_lex, asta_delpv, allok, sempar, phonpar, neologisme]

thefunctions = thetarspfunctions + thestapfunctions + theastafunctions

str2functionmap: Dict[str, QueryFunction] = {}

for f in thefunctions:
    fname = getfname(f)
    str2functionmap[fname] = f

junk = 0

# Used by SASTA to find form functions
form_map: Dict[str, Callable] = {
    'TARSP': mktarspform,
    'ASTA': astaform,
    'STAP': makestapform
}



'''
The module *smallclauses* deals with *small clauses*, i.e. utterances with a
predication but no finite verb. Examples include:

* die dicht
* schoenen aan
* ladder hier
* ik varken
* nou Marije nat
* traktor niet hier
* mama even drinken
* nee dees wel
* ikke dees pakke
* ik naar omie

Alpino cannot analyse such utterances in a proper manner. For example,
it analyses *schoenen aan* as a sequence of two discourse part, one a noun, the other an
adjective.

For this reason this module inserts an appropriate finite verb into these utterances. This modified utterance with
a finite verb can be analysed by Alpino. In the resulting parse, SASTA removes the verb
again and the resulting parse tree is subjected to the queries for analysis.

For example, by inserting the finite verb *moeten* in the utterance *schoenen aan*,
resuting in *schoenen moeten aan*, *schoenen* can be analysed as a subject noun and *aan* as predicative adjective,
and the whole utterance as an instance of TARSP *OndVC*.

The small clause utterances are often very short, and we analysed utterances with 2
and 3 "real" words,  "real" in the sense that we excluded interjections, filled pauses,
interpunction signs, CHAT codes such as *xxx*, etc.

The different subcases are dealt with by the function *smallclauses*:

.. autofunction:: smallclauses::smallclauses

'''

from typing import List

from sastadev.conf import settings
from sastadev.dedup import filledpauseslexicon
from sastadev.lexicon import known_word, tswnouns
from sastadev.metadata import (SASTA, Meta, bpl_delete, bpl_none,
                               defaultpenalty, insertion,
                               insertiontokenmapping, smallclause,
                               tokenmapping)
from sastadev.namepartlexicon import namepart_isa_namepart
from sastadev.sastatoken import Token
from sastadev.sastatypes import SynTree
from sastadev.tokenmd import TokenListMD
from sastadev.top3000 import (genlexicon, intransitive, isanimate, ishuman,
                              pseudotr, transitive)
from sastadev.treebankfunctions import getattval, getnodeyield

space = ' '
biglocvzs = ['achter', 'beneden', 'binnen', 'boven', 'bovenop', 'buiten', 'dichtbij']
#surenouns = ['mama', 'papa'] replaced by tswnouns from lexicon
longvowels = ['a', 'é', 'i', 'o', 'u', 'y']
vowels = ['a', 'e', 'i', 'o', 'u']

uniquelynominativeperspros = ['ik', 'jij', 'hij', 'zij', 'wij', 'ikke', "'k", "k", "ie", "we"]


def makegen(lemma):
    if lemma is None or len(lemma) < 2:
        result = None
    elif lemma[-1] in ['s', 'z', 'x']:
        result = lemma + "'"
    elif lemma[-2:] in ['ij']:
        result = lemma + 's'
    elif lemma[-2] in vowels and lemma[-1] in vowels:
        result = lemma + 's'
    elif lemma[-1] in longvowels:
        result = lemma + "'s"
    else:
        result = lemma + 's'
    return result


def realword(node):
    result = True
    result = result and getattval(node, 'pt') not in ['tsw', 'let']
    result = result and getattval(node, 'lemma') not in ['xx', 'xxx', 'yyy', 'www', 'hè']
    result = result and getattval(node, 'lemma') not in filledpauseslexicon
    result = result or lemma(node) in tswnouns

    return result


def hasgenitive(node):
    lemma = getattval(node, 'lemma')
    nodept = pt(node)
    if nodept not in ['n', 'vnw']:
        nodept = 'n'
    result = (lemma, nodept) in genlexicon and 'yes' in genlexicon[(lemma, nodept)]
    result = result or namepart_isa_namepart(lemma)
    return result


def aanwvnw(node):
    result = getattval(node, 'pt') == 'vnw' and getattval(node, 'vwtype') == 'aanw' and not rpronoun(node)
    return result


def n(node):
    result = getattval(node, 'pt') == 'n'
    return result


def getal(node):
    result = getattval(node, 'getal')
    return result


def pt(node):
    result = getattval(node, 'pt')
    return result


def bg(node):
    result = int(getattval(node, 'begin'))
    return result


def tw(node):
    result = getattval(node, 'pt') == 'tw'
    return result


def word(node):
    result = getattval(node, 'word')
    return result


def adj(node):
    result = getattval(node, 'pt') == 'adj'
    return result


def perspro(node):
    pt = getattval(node, 'pt')
    vwtype = getattval(node, 'vwtype')
    result = pt == 'vnw' and vwtype == 'pers'
    return result


def nomperspro(node):
    lemma = getattval(node, 'lemma')
    result = perspro(node) and lemma in uniquelynominativeperspros
    return result


def inf(node):
    result = getattval(node, 'pt') == 'ww' and getattval(node, 'wvorm') == 'inf'
    return result


def rpronoun(node):
    result = getattval(node, 'pt') == 'vnw' and \
        getattval(node, 'lemma') in ['er', 'hier', 'daar', 'ergens', 'overal', 'nergens', 'waar']
    return result


def bw(node):
    result = getattval(node, 'pt') == 'bw'
    return result


def ww(node):
    result = getattval(node, 'pt') == 'ww'
    return result


def lemma(node):
    result = getattval(node, 'lemma')
    return result


def predadv(node):
    result = locadv(node)
    result = result or (bw(node) and lemma(node) in ['niet', 'mee', 'weg'])
    return result


def vz(node):
    result = getattval(node, 'pt') == 'vz'
    return result


def locadv(node):
    result = getattval(node, 'pt') in ['bw', 'vz']
    frame = getattval(node, 'frame')
    result = result and ('loc' in frame or 'er_adverb' in frame)
    result = result or rpronoun(node)
    return result


def biglocvz(node):
    result = getattval(node, 'lemma') in biglocvzs
    return result


def istswnoun(node):
    result = getattval(node, 'lemma') in tswnouns
    return result


def getleavestr(leaves):
    leaveseq = ['{}:{}:{}:{}'.format(getattval(leave, 'end'), getattval(leave, 'word'), getattval(leave, 'lemma'),
                                     getattval(leave, 'pt')) for leave
                in leaves]
    leavestr = space.join(leaveseq)
    return leavestr


def knownnoun(node):
    word = getattval(node, 'word')
    lemma = getattval(node, 'lemma')
    postag = pt(node)
    result = postag == 'n' and (known_word(word) or known_word(lemma))
    result = result or lemma in tswnouns
    return result


def nominal(node):
    result = pt(node) == 'n' or aanwvnw(node)
    return result


def mktoken(node, map):
    nodebegin = bg(node)
    nodeword = word(node)
    if nodebegin in map:
        nodepos = map[nodebegin]
    else:
        settings.LOGGER.error('missing begin in map {}'.format(nodebegin))
        nodepos = int(nodebegin)
    result = Token(nodeword, nodepos)
    return result


def mktokenlist(tokens, fpos, inserttokens):
    resultlist = [token for token in tokens if token.pos <= fpos] + \
        inserttokens + \
                 [token for token in tokens if token.pos > fpos]
    return resultlist


def oldmktokenlist(leaves, themap, fpos, inserttokens):
    resultlist = [mktoken(lv, themap) for lv in leaves if bg(lv) <= fpos] + \
        inserttokens + \
                 [mktoken(lv, themap) for lv in leaves if bg(lv) > fpos]
    return resultlist


def mkinsertmeta(inserttokens, resultlist):
    insertposs = [token.pos + token.subpos for token in inserttokens]
    insertwordlist = [token.word for token in inserttokens]
    tokenmappinglist = [token.pos if token.subpos == 0 else None for token in resultlist]
    metadata1 = [Meta(insertion, [insertword], annotatedposlist=[insertpos],
                 annotatedwordlist=[], annotationposlist=[insertpos],
                 annotationwordlist=[insertword], cat=smallclause, source=SASTA, penalty=defaultpenalty,
                 backplacement=bpl_delete) for insertword, insertpos in zip(insertwordlist, insertposs)]
    meta2 = Meta(insertiontokenmapping, tokenmappinglist, cat=tokenmapping, source=SASTA, penalty=0,
                 backplacement=bpl_none)
    metadata = metadata1 + [meta2]
    return metadata


def smallclauses(tokensmd: TokenListMD, tree: SynTree) -> List[TokenListMD]:
    '''

    :param tokensmd: list of tokens with metadata
    :param tree: the syntactic structure of the utterance
    :return: a possibly empty list of alternative TokenListMD objects

    The function *smallclauses* creates zero or more alternative TokenListMD
    objects by inserting an appropriate finite verb.

    "Appropriate"in this comtext means that it must fit in *syntactically* and
    *morphologically* in this context:

    * **syntactically** the verb must take the right complements
    * **morphologically** the verb form must agree with the subject

    SASTA usually inserts a verb that has few different inflectional forms, so that the
    agreement requirement is met in the easiest way. (e.g. *moeten* has only one form
    in present tense for singular (*moet*), and only one form for plural (*moeten*).

    The relevant cases from the available TARSP-examples have been inventoried,
    and some of the relevant cases have been implemented. The table below specifies which cases are currently covered.

    Cases covered so far:

    .. csv-table:: Small Clause Subcases covered
      :file: Documentation/smallclausetable.csv
      :widths: 16, 16, 16, 16, 16, 20
      :header-rows: 1


    '''
    resultlist = []
    leaves = getnodeyield(tree)
    reducedleaves = [leave for leave in leaves if realword(leave)]
    if not (len(reducedleaves) > 1 and len(reducedleaves) <= 3):
        return resultlist
    tokens = tokensmd.tokens
    treewords = [word(tokennode) for tokennode in leaves]
    tokenwords = [token.word for token in tokens if not token.skip]
    if treewords != tokenwords:
        settings.LOGGER.error('Token mismatch: {} v. {}'.format(treewords, tokenwords))
        return []
    themap = {bg(tokennode): token.pos for (tokennode, token) in zip(leaves, tokens)}
    metadata = tokensmd.metadata

    if len(reducedleaves) <= 3:
        first = leaves[0]
        second = leaves[1]
    if len(reducedleaves) == 3:
        third = leaves[0]

    if len(reducedleaves) == 2:
        if (aanwvnw(first) or knownnoun(first) or perspro(first)) and (predadv(second) or vz(second) or bw(second)):
            fpos = int(getattval(first, 'begin'))
            inserttokens = [Token('moet' if getal(first) != 'mv' else 'moeten', fpos, subpos=5)]
            resultlist = mktokenlist(tokens, fpos, inserttokens)
            metadata += mkinsertmeta(inserttokens, resultlist)
        #elif (aanwvnw(second) or knownnoun(second) or perspro(second) or tw(second)) and predadv(first):
        elif nomperspro(second) and predadv(first):
            fpos = int(getattval(first, 'begin'))
            inserttokens = [Token('moet' if getal(second) != 'mv' else 'moeten', fpos, subpos=5)]
            resultlist = mktokenlist(tokens, fpos, inserttokens)
            metadata += mkinsertmeta(inserttokens, resultlist)
        elif (aanwvnw(first) or knownnoun(first)) and adj(second):
            fpos = int(getattval(first, 'begin'))
            inserttokens = [Token('is' if getal(first) != 'mv' else 'zijn', fpos, subpos=5)]
            resultlist = mktokenlist(tokens, fpos, inserttokens)
            metadata += mkinsertmeta(inserttokens, resultlist)
        elif (aanwvnw(second) or knownnoun(second) or tw(second)) and biglocvz(first):
            fpos = int(getattval(first, 'begin'))
            inserttokens = [Token('is' if getal(first) != 'mv' else 'zijn', fpos, subpos=5)]
            resultlist = mktokenlist(tokens, fpos, inserttokens)
        elif knownnoun(first) and knownnoun(second) and not (lemma(first) == lemma(second)):
            if hasgenitive(first):
                genform = makegen(lemma(first))
                fpos = int(getattval(first, 'begin'))
                inserttokens = [Token("z'n", fpos, subpos=5)]
                resultlist = mktokenlist(tokens, fpos, inserttokens)
                metadata += mkinsertmeta(inserttokens, resultlist)
            else:
                fpos = int(getattval(first, 'begin'))
                inserttokens = [Token('is' if getal(first) != 'mv' else 'zijn', fpos, subpos=5)]
                resultlist = mktokenlist(tokens, fpos, inserttokens)
                metadata += mkinsertmeta(inserttokens, resultlist)
        elif (aanwvnw(first) or knownnoun(first) or istswnoun(first)) and inf(second):
            if intransitive(second):
                firstsubject = True
            elif transitive(second) and ishuman(first):
                firstsubject = True
            elif pseudotr(second) and (ishuman(first) or isanimate(first)):
                firstsubject = True
            else:
                firstsubject = False
            if firstsubject:
                fpos = int(getattval(first, 'begin'))
                inserttokens = [Token('wil' if getal(first) != 'mv' else 'willen', fpos, subpos=5)]
            else:
                fpos = -1
                inserttokens = [Token('ik', fpos, subpos=5), Token('wil', fpos, subpos=8)]
            resultlist = mktokenlist(tokens, fpos, inserttokens)
            metadata += mkinsertmeta(inserttokens, resultlist)
        elif not nominal(first) and not ww(first) and inf(second):
            fpos = -1
            inserttokens = [Token('ik', fpos, subpos=5), Token('wil', fpos, subpos=8)]
            resultlist = mktokenlist(tokens, fpos, inserttokens)
            metadata += mkinsertmeta(inserttokens, resultlist)
    if resultlist == []:
        result = []
    else:
        result = [TokenListMD(resultlist, metadata)]
    return result



from lxml import etree

streexml = {}
stree = {}

streexml[1] = '''
<alpino_ds version="1.3" id="Tarvb2-22.xml">
  <node begin="0" cat="top" end="8" id="0" rel="top" highlight="yes">
    <node begin="0" cat="whq" end="7" id="1" rel="--" highlight="yes">
      <node begin="0" case="both" def="indef" end="1" frame="pronoun(ywh,thi,sg,het,both,indef,nparg)" gen="het" getal="ev" id="2" index="1" lcat="np" lemma="wat" naamval="stan" num="sg" pdtype="pron" per="thi" persoon="3o" pos="pron" postag="VNW(vb,pron,stan,vol,3o,ev)" pt="vnw" rel="whd" rnum="sg" root="wat" sense="wat" special="nparg" status="vol" vwtype="vb" wh="ywh" word="wat" highlight="yes"/>
      <node begin="0" cat="sv1" end="7" id="3" rel="body" highlight="yes">
        <node begin="1" end="2" frame="verb('hebben/zijn',pl,modifier(aux(inf)))" id="4" infl="pl" lcat="sv1" lemma="moeten" pos="verb" postag="WW(pv,tgw,mv)" pt="ww" pvagr="mv" pvtijd="tgw" rel="hd" root="moet" sc="modifier(aux(inf))" sense="moet" stype="whquestion" tense="present" word="moeten" wvorm="pv" highlight="yes"/>
        <node begin="2" case="nom" def="def" end="3" frame="pronoun(nwh,fir,pl,de,nom,def,wkpro)" gen="de" getal="mv" id="5" index="2" lcat="np" lemma="we" naamval="nomin" num="pl" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,red,1,mv)" pt="vnw" rel="su" rnum="pl" root="we" sense="we" special="wkpro" status="red" vwtype="pers" wh="nwh" word="we" highlight="yes"/>
        <node begin="0" cat="inf" end="7" id="6" rel="vc" highlight="yes">
          <node begin="0" end="1" id="7" index="1" rel="obj1" highlight="yes"/>
          <node begin="2" end="3" id="8" index="2" rel="su" highlight="yes"/>
          <node begin="3" cat="pp" end="6" id="9" rel="mod" highlight="yes">
            <node begin="3" end="4" frame="preposition(met,[mee,[en,al]])" id="10" lcat="pp" lemma="met" pos="prep" postag="VZ(init)" pt="vz" rel="hd" root="met" sense="met" vztype="init" word="met" highlight="yes"/>
            <node begin="4" cat="np" end="6" id="11" rel="obj1" highlight="yes">
              <node begin="4" buiging="zonder" end="5" frame="determiner(de,nwh,nmod,pro,nparg)" id="12" infl="de" lcat="detp" lemma="die" naamval="stan" npagr="rest" pdtype="det" pos="det" positie="prenom" postag="VNW(aanw,det,stan,prenom,zonder,rest)" pt="vnw" rel="det" root="die" sense="die" vwtype="aanw" wh="nwh" word="die" highlight="yes"/>
              <node begin="5" end="6" frame="noun(de,count,sg)" gen="de" genus="zijd" getal="ev" graad="basis" id="13" lcat="np" lemma="stuk_kies" naamval="stan" ntype="soort" num="sg" pos="noun" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="hd" rnum="sg" root="stuk_kies" sense="stuk_kies" word="stukkies" highlight="yes"/>
            </node>
          </node>
          <node begin="6" buiging="zonder" end="7" frame="verb(hebben,inf(no_e),transitive_ndev)" id="14" infl="inf(no_e)" lcat="inf" lemma="doen" pos="verb" positie="vrij" postag="WW(inf,vrij,zonder)" pt="ww" rel="hd" root="doe" sc="transitive_ndev" sense="doe" word="doen" wvorm="inf" highlight="yes"/>
        </node>
      </node>
    </node>
    <node begin="7" end="8" frame="punct(vraag)" id="15" lcat="punct" lemma="?" pos="punct" postag="LET()" pt="let" rel="--" root="?" sense="?" special="vraag" word="?" highlight="yes"/>
  </node>
  <sentence sentid="22 ">wat moeten we met die stukkies doen ?</sentence>
  <metadata/>
</alpino_ds>
'''


streexml[2] = '''
<alpino_ds version="1.6" id="ASTA_06-ASTA_sample_06_001.xml">
  <parser cats="2" skips="0"/>
  <node begin="0" cat="top" end="6" id="0" rel="top" highlight="yes">
    <node begin="0" cat="du" end="6" id="1" rel="--" highlight="yes">
      <node begin="0" cat="smain" end="2" id="2" rel="dp" highlight="yes">
        <node begin="0" end="1" frame="determiner(het,nwh,nmod,pro,nparg)" getal="ev" his="normal" his_1="decap" his_1_1="normal" id="3" infl="het" lcat="np" lemma="dat" naamval="stan" pdtype="pron" persoon="3o" pos="det" postag="VNW(aanw,pron,stan,vol,3o,ev)" pt="vnw" rel="su" rnum="sg" root="dat" sense="dat" status="vol" vwtype="aanw" wh="nwh" word="Dat" highlight="yes"/>
        <node begin="1" end="2" frame="verb(unacc,sg_heeft,intransitive)" his="normal" his_1="normal" id="4" infl="sg_heeft" lcat="smain" lemma="zijn" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="ben" sc="intransitive" sense="ben" stype="declarative" tense="present" word="is" wvorm="pv" highlight="yes"/>
      </node>
      <node begin="2" cat="smain" end="6" id="5" rel="dp" highlight="yes">
        <node begin="2" end="3" frame="determiner(het,nwh,nmod,pro,nparg)" getal="ev" his="normal" his_1="normal" id="6" infl="het" lcat="np" lemma="dat" naamval="stan" pdtype="pron" persoon="3o" pos="det" postag="VNW(aanw,pron,stan,vol,3o,ev)" pt="vnw" rel="obj1" rnum="sg" root="dat" sense="dat" status="vol" vwtype="aanw" wh="nwh" word="dat" highlight="yes"/>
        <node begin="3" end="4" frame="verb(hebben,sg,transitive)" his="normal" his_1="normal" id="7" infl="sg" lcat="smain" lemma="weten" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="weet" sc="transitive" sense="weet" stype="declarative" tense="present" word="weet" wvorm="pv" highlight="yes"/>
        <node begin="4" case="nom" def="def" end="5" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" his="normal" his_1="normal" id="8" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik" highlight="yes"/>
        <node begin="5" end="6" frame="adverb" his="normal" his_1="normal" id="9" lcat="advp" lemma="niet" pos="adv" postag="BW()" pt="bw" rel="mod" root="niet" sense="niet" word="niet" highlight="yes"/>
      </node>
    </node>
  </node>
  <sentence sentid="1">Dat is dat weet ik niet</sentence>
  <metadata>
    <meta type="text" name="charencoding" value="UTF8"/>
    <meta type="text" name="childage" value=""/>
    <meta type="text" name="childmonths" value=""/>
    <meta type="text" name="comment" value="##META text samplenaam = ASTA-06"/>
    <meta type="text" name="session" value="ASTA_sample_06"/>
    <meta type="text" name="origutt" value="Dat is dat weet ik niet"/>
    <meta type="text" name="parsefile" value="Unknown_corpus_ASTA_sample_06_u00000000002.xml"/>
    <meta type="text" name="speaker" value="PMA"/>
    <meta type="int" name="uttendlineno" value="10"/>
    <meta type="int" name="uttid" value="1"/>
    <meta type="int" name="uttstartlineno" value="10"/>
    <meta type="text" name="name" value="pma"/>
    <meta type="text" name="SES" value=""/>
    <meta type="text" name="age" value=""/>
    <meta type="text" name="custom" value=""/>
    <meta type="text" name="education" value=""/>
    <meta type="text" name="group" value=""/>
    <meta type="text" name="language" value="nld"/>
    <meta type="text" name="months" value=""/>
    <meta type="text" name="role" value="Other"/>
    <meta type="text" name="sex" value=""/>
    <meta type="text" name="xsid" value="1"/>
    <meta type="int" name="uttno" value="2"/>
  </metadata>
</alpino_ds>
'''

for el in streexml:
    stree[el] = etree.fromstring(streexml[el])



import os

from sastadev.constants import formsfolder, intreebanksfolder

formext = '.xlsx'


def getformfilename(infullname, suffix):
    base, infilename = os.path.split(infullname)
    (filebase, ext) = os.path.splitext(infilename)
    (core, lastfolder) = os.path.split(base)
    if lastfolder == intreebanksfolder:
        formname = os.path.join(core, formsfolder, filebase + suffix + formext)
    else:
        formname = base + suffix + formext
    return formname



import csv

from sastadev.conf import settings

tab = '\t'
mysep = tab


def readcsv(filename, sep=mysep, header=True, quotechar='"', encoding='utf8'):
    result = []
    try:
        infile = open(filename, 'r', encoding=encoding, newline='')
    except FileNotFoundError as e:
        settings.LOGGER.error(e)
        return result
    myreader = csv.reader(infile, delimiter=sep, quotechar=quotechar)
    # read the header if needed
    if header:
        next(myreader)
    rowctr = 1
    for row in myreader:
        result.append((rowctr, row))
        rowctr += 1
    infile.close()
    return result


def readheadedcsv(filename, sep=mysep, quotechar='"', encoding='utf8'):
    result = []
    header = []
    try:
        infile = open(filename, 'r', encoding=encoding, newline='')
    except FileNotFoundError as e:
        settings.LOGGER.warning(e)
        return header, result
    myreader = csv.reader(infile, delimiter=sep, quotechar=quotechar)
    # read the header
    try:
        header = next(myreader)
    except StopIteration as e:
        settings.LOGGER.warning(e)
        pass
    rowctr = 1
    for row in myreader:
        result.append((rowctr, row))
        rowctr += 1
    infile.close()
    return (header, result)


def writecsv(rows, filename, header=[], sep=tab, quotechar='"'):
    try:
        outfile = open(filename, 'w', encoding='utf8', newline='')
    except FileExistsError:
        pass
    mywriter = csv.writer(outfile, delimiter=sep, quotechar=quotechar, quoting=csv.QUOTE_MINIMAL)
    if header != []:
        mywriter.writerow(header)
    for row in rows:
        mywriter.writerow(row)
    outfile.close()



from sastadev.conf import settings

pre_process_str, core_process_str, post_process_str, form_process_str = 'pre', 'core', 'post', 'form'
pre_process, core_process, post_process, form_process = 0, 1, 2, 3


def getprocess(process):
    if process.lower() == core_process_str:
        result = core_process
    elif process.lower() == post_process_str:
        result = post_process
    elif process.lower() == pre_process_str:
        result = pre_process
    elif process.lower() == form_process_str:
        result = form_process
    else:
        result = -1
        settings.LOGGER.error('Illegal value for process {}'.format(process))
    return result


def clean(valstr):
    result = valstr.strip().lower()
    return result


class Query:
    def __init__(self, id, cat, subcat, level, item, altitems, implies, original, pages, fase, query, inform,
                 screening, process, special1, special2, comments):
        self.id = id
        self.cat = cat
        self.subcat = subcat
        self.level = level
        self.item = item
        self.altitems = altitems
        self.implies = implies
        self.original = original
        self.pages = pages
        self.fase = fase
        self.query = query
        self.inform = inform
        self.screening = screening
        self.process = getprocess(process)
        self.special1 = clean(special1)
        self.special2 = special2
        self.comments = comments


def query_inform(query):
    result = query.inform == "yes"
    return result


def is_preorcore(query):
    result = is_pre(query) or is_core(query)
    return result


def is_pre(query):
    result = (query.process == pre_process)
    return result


def is_core(query):
    result = (query.process == core_process)
    return result


def is_post(query):
    result = (query.process == post_process)
    return result


def query_exists(query):
    result = query.query != "" and query.query is not None
    return result



import xlsxwriter
from openpyxl import load_workbook

from sastadev.conf import settings

eps = ''
defaultcolwidth = 15
cola, colb, colc, cold, cole, colf, colg, colh, coli, colj, colk, coll, colm, coln, colo, colp, colq, colr, cols, \
    colt, colu, colv, colw, colx, coly, colz = 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, \
    21, 22, 23, 24, 25


def xlsx_writerow(sheet, rowctr, row, format=None, formats=[]):
    lrow = len(row)
    lformats = len(formats)
    for colctr in range(lrow):
        if format is None:
            if formats != [] and colctr < lformats and formats[colctr] is not None:
                sheet.write(rowctr, colctr, row[colctr], formats[colctr])
            else:
                sheet.write(rowctr, colctr, row[colctr])
        else:
            sheet.write(rowctr, colctr, row[colctr], format)


def writetable(tabel, ws, startrow=0, startcol=0, rhformat=None, chformat=None, cellformat=None):
    curcol = startcol
    currow = startrow
    for row in tabel:
        for el in row:
            if currow == startrow:
                theformat = rhformat
            elif curcol == startcol:
                theformat = chformat
            else:
                theformat = cellformat
            ws.write(currow, curcol, el, theformat)
            curcol += 1
        currow += 1
        curcol = startcol


def mkworkbook(outfullname, headers, allrows, sheetname='Sheet1', freeze_panes=None, formats=[]):
    workbook = xlsxwriter.Workbook(outfullname, {"strings_to_numbers": True})
    bold = workbook.add_format({'bold': True})

    realformats = []
    for fmt in formats:
        if fmt is None:
            realformats.append(fmt)
        else:
            realfmt = workbook.add_format(fmt)
            realformats.append(realfmt)

    worksheet1 = workbook.add_worksheet(sheetname)

    # worksheet1
    if freeze_panes is not None:
        (r, c) = freeze_panes
        worksheet1.freeze_panes(r, c)

    colctr = 0
    if headers != []:
        for val in headers[-1]:
            if val is None:
                cval = ''
            else:
                cval = str(val)
            colwidth = len(cval) if len(cval) > defaultcolwidth else defaultcolwidth
            worksheet1.set_column(colctr, colctr, colwidth)
            colctr += 1

    rowctr = 0
    for header in headers:
        xlsx_writerow(worksheet1, rowctr, header, format=bold)
        rowctr += 1

    for row in allrows:
        xlsx_writerow(worksheet1, rowctr, row, formats=realformats)
        rowctr += 1

    worksheet1.autofilter(0, 0, rowctr, colctr)
    return workbook


def adaptformats(formats, workbook):
    realformats = []
    for fmt in formats:
        if fmt is None:
            realformats.append(fmt)
        else:
            realfmt = workbook.add_format(fmt)
            realformats.append(realfmt)
    return realformats


def add_worksheet(workbook, headers, allrows, sheetname='Sheet2', freeze_panes=None, formats=[]):
    bold = workbook.add_format({'bold': True})

    realformats = adaptformats(formats, workbook)

    worksheet1 = workbook.add_worksheet(sheetname)

    # worksheet1
    if freeze_panes is not None:
        (r, c) = freeze_panes
        worksheet1.freeze_panes(r, c)

    if headers != []:
        colctr = 0
        for val in headers[-1]:
            if val is None:
                cval = ''
            else:
                cval = str(val)
            colwidth = len(cval) if len(cval) > defaultcolwidth else defaultcolwidth
            worksheet1.set_column(colctr, colctr, colwidth)
            colctr += 1

    rowctr = 0
    for header in headers:
        xlsx_writerow(worksheet1, rowctr, header, format=bold)
        rowctr += 1

    for row in allrows:
        xlsx_writerow(worksheet1, rowctr, row, formats=realformats)
        rowctr += 1

    worksheet1.autofilter(0, 0, rowctr, colctr)
    return worksheet1


def getxlsxdata(fullname, headerrow=0, sheetname=None):
    data = []
    header = []
    try:
        wb = load_workbook(fullname)
    except Exception:
        return (header, data)
    wsns = wb.sheetnames
    if wsns != []:
        if sheetname is None:
            ws = wb[wsns[0]]
        else:
            ws = wb[sheetname]
    else:
        settings.LOGGER.error("Error reading {}. No worksheets found\n".format(fullname))
    rowctr = -1
    for row in ws.iter_rows():
        rowctr += 1
        #preheader
        if rowctr < headerrow:
            continue
        # header
        elif rowctr == headerrow:
            header = [eps if cell.value is None else cell.value for cell in row]
        else:
            valuerow = [eps if cell.value is None else cell.value for cell in row]
            data.append(valuerow)
    wb.close()
    return header, data


def write2excel(datadict, header, filename):
    data = [list(key) + datadict[key] for key in datadict]
    workbook = mkworkbook(filename, [header], data)
    workbook.close()



'''
defines functions for the STAP post part of the methods

'''
from collections import Counter

BB_ids = ['S010', 'S011', 'S012']


def BB_totaal(allresults, _):
    scores = []
    for qid in BB_ids:
        if qid in allresults.coreresults:
            scores.append(allresults.coreresults[qid])
        else:
            scores.append(Counter())
    counts = [len(s) for s in scores]
    result = sum(counts)
    return result


def GLVU(allresults, _):
    total_length_VU = 0
    if 'S013' in allresults.coreresults:
        for key in allresults.coreresults['S013']:
            total_length_VU += allresults.coreresults['S013'][key]
    result = total_length_VU / allresults.uttcount
    return result


def GL5LVU(allresults, _):
    counts = []
    if 'S013' in allresults.coreresults:
        for key in allresults.coreresults['S013']:
            counts.append(allresults.coreresults['S013'][key])
    sorted_counts = counts.sort()
    result = sum(counts[45:50]) / 5
    return result



space = ' '


class Token:
    def __init__(self, word, pos, skip=False, subpos=0):
        self.word = word
        self.pos = pos
        self.subpos = subpos
        self.skip = skip

    def __repr__(self):
        fmtstr = 'Token(word={},pos={}, skip={}, subpos={})'
        result = fmtstr.format(repr(self.word), repr(self.pos), repr(self.skip), repr(self.subpos))
        return result

    def __str__(self):
        skipstr = ' (skip=True)' if self.skip else ''
        subposstr = '/{}'.format(self.subpos) if self.subpos != 0 else ''
        result = '{}{}:{}{}'.format(self.pos, subposstr, self.word, skipstr)
        return result


def oldstringlist2tokenlist(list):
    result = []
    llist = len(list)
    for el in range(llist):
        thetoken = Token(list[el], el)
        result.append(thetoken)
    return result


def stringlist2tokenlist(list, start=0, inc=1):
    result = []
    llist = len(list)
    pos = start
    for el in range(llist):
        thetoken = Token(list[el], pos)
        result.append(thetoken)
        pos += inc
    return result


def tokenlist2stringlist(tlist, skip=False):
    if skip:
        result = [t.word for t in tlist if not t.skip]
    else:
        result = [t.word for t in tlist]
    return result


def tokenlist2string(tlist):
    wordlist = [t.word for t in tlist]
    result = space.join(wordlist)
    return result


def show(tokenlist):
    resultlist = []
    for token in tokenlist:
        resultlist.append(str(token))
    result = ', '.join(resultlist)
    return result


def tokeninflate(token):
    result = inflate(token.pos) + token.subpos
    return result


def deflate(n: int):
    result = (n // 10) - 1
    return result


def inflate(n: int):
    result = (n + 1) * 10
    return result


def insertinflate(n: int):
    dm = n % 10
    result = ((n - dm) + 1) * 10 + dm
    return result



from typing import List

from lxml import etree

from sastadev.conf import settings
from sastadev.lexicon import (getinflforms, getwordposinfo, informlexiconpos,
                              pvinfl2dcoi)
from sastadev.metadata import bpl_node, mkSASTAMeta
from sastadev.sastatoken import Token
from sastadev.sastatypes import SynTree, UttId
from sastadev.tokenmd import TokenListMD
from sastadev.treebankfunctions import (copymodifynode, find1, getattval,
                                        getdetof, getheadof, getlemma,
                                        indextransform, inverted, lbrother,
                                        nominal, rbrother, showtree,
                                        simpleshow)

debug = False

nominalpts = ['n', 'vnw', 'tw']
nominalisablepts = ['adj', 'ww']

#: The constant *normalsentencexpath* defines a query for a finite verb form
#: with *smain* or *sv1* as parent node that is the only phrase under the top node,
#: possibly accompanied by punctuation signs.
normalsentencexpath = """.//node[@pt="ww" and  @wvorm="pv" and
       parent::node[(@cat="smain" or @cat="sv1") and
           parent::node[@cat="top" and count(node[@cat or  @pt!="let"])=1]]]"""

#: The constant *normalwhqsentencexpath* defines a query for a finite verb form with as parent a node with
#: *cat*=*sv1* which has a parent node with *cat*=*whq* that is the only phrase under the top node,
#: possibly accompanied by punctuation signs.
normalwhqsentencexpath = """
//node[@pt="ww" and  @wvorm="pv" and
       parent::node[( @cat="sv1") and
           parent::node[@cat="whq" and
               parent::node[@cat="top" and count(node[@cat or  @pt!="let"])=1]]]]"""
#: The constant *abnormalobj2sentencexpath* defines a query for a finite verb of which the parent contains
#: as indirect object (*obj2*) a pronoun or  the word *zij*. (*zij* can also be analysed, usually incorrectly
#: in the SASTA context, as the subjunctive form of the verb *zijn*).
#: example sentence to which this applies: 	*mij lukt het ook*
abnormalobj2sentencexpath = """.//node[@pt="ww" and  @wvorm="pv" and
       parent::node[node[@rel="obj2" and (@pt="vnw" or @word="zij")]]]
"""
#: The constant *abnormalobj2xpath* searches for nodes that are an indirect object and either a pronoun or the word "zij"
#: and that have a sibling finite verb. @I do not understand this anymore@
abnormalobj2xpath = """.//node[@rel="obj2" and (@pt="vnw" or @word="zij")
        and parent::node[node[@pt="ww" and @wvorm="pv"]]]"""

#: The constant *subjxpath* defines a query to search for a subject that is a sibling of a finite verb.
subjxpath = """.//node[@rel="su" and parent::node[node[@pt="ww" and @wvorm="pv"]]]"""

zijsgnodestringtemplate = """
        <node begin="{begin}" case="nom" def="def" end="{end}" frame="pronoun(nwh,thi,both,de,nom,def)" gen="de"
        genus="fem" getal="ev"  lcat="np" lemma="zij" naamval="nomin" num="both" pdtype="pron"
        per="thi" persoon="3v" pos="pron" postag="VNW(pers,pron,nomin,vol,3v,ev,fem)" pt="vnw"
        rel="{rel}" rnum="sg" root="zij" sense="zij" status="vol" vwtype="pers" wh="nwh" word="zij"/>
"""

zesgnodestringtemplate = """
        <node begin="{begin}" case="both" def="def" end="{end}" frame="pronoun(nwh,thi,both,de,both,def,wkpro)" gen="de"
        genus="fem" getal="ev"  lcat="np" lemma="ze" naamval="stan" num="both" pdtype="pron" per="thi" persoon="3"
        pos="pron" postag="VNW(pers,pron,stan,red,3,ev,fem)" pt="vnw" rel="{rel}" rnum="sg" root="ze"
        sense="ze" special="wkpro" status="red" vwtype="pers" wh="nwh" word="ze"/>
"""

zeplnodestringtemplate = """

      <node begin="{begin}" case="both" def="def" end="{end}" frame="pronoun(nwh,thi,both,de,both,def,wkpro)" gen="de"
      getal="mv"  lcat="--" lemma="ze" naamval="stan" num="both" pdtype="pron" per="thi" persoon="3" pos="pron"
      postag="VNW(pers,pron,stan,red,3,mv)" pt="vnw" rel="{rel}" root="ze" sense="ze"
      special="wkpro" status="red" vwtype="pers" wh="nwh" word="ze"/>

"""


jenodestringtemplate = """<node begin="{begin}" case="both" def="def" end="{end}"
frame="pronoun(nwh,je,sg,de,both,def,wkpro)" gen="de" getal="ev"  lcat="np" lemma="je" naamval="nomin"
num="sg" pdtype="pron" per="je" persoon="2v" pos="pron" postag="VNW(pers,pron,nomin,red,2v,ev)"
pt="vnw" rel="{rel}" rnum="sg" root="je" sense="je" special="wkpro" status="red" vwtype="pers" wh="nwh" word="je"/>
"""

hetnodestringtemplate = """<node begin="{begin}" end="{end}" frame="determiner(het,nwh,nmod,pro,nparg,wkpro)"
genus="onz" getal="ev"  infl="het" lcat="np" lemma="het" naamval="stan" pdtype="pron" persoon="3"
pos="det" postag="VNW(pers,pron,stan,red,3,ev,onz)" pt="vnw" rel="{rel}" rnum="sg" root="het"
sense="het" status="red" vwtype="pers" wh="nwh" word="het"/>
"""

tnodestringtemplate = """<node begin="{begin}" end="{end}" frame="determiner(het,nwh,nmod,pro,nparg,wkpro)"
genus="onz" getal="ev"  infl="het" lcat="np" lemma="het" naamval="stan" pdtype="pron" persoon="3" pos="det"
postag="VNW(pers,pron,stan,red,3,ev,onz)" pt="vnw" rel="{rel}" rnum="sg" root="het" sense="het"
status="red" vwtype="pers" wh="nwh" word="'t"/>
"""

usgnodestringtemplate = """
        <node begin="{begin}" case="both" def="def" end="{end}" frame="pronoun(nwh,u,sg,de,both,def)"
        gen="de" getal="ev"  lcat="np" lemma="u" naamval="nomin" num="sg" pdtype="pron" per="u" persoon="2b"
        pos="pron" postag="VNW(pers,pron,nomin,vol,2b,ev)" pt="vnw" rel="{rel}" rnum="sg" root="u" sense="u"
        status="vol" vwtype="pers" wh="nwh" word="u"/>
"""

jullienodestringtemplate = """
        <node begin="{begin}" case="both" def="def" end="{end}" frame="pronoun(nwh,je,pl,de,both,def)" gen="de"
        getal="mv"  lcat="np" lemma="jullie" naamval="stan" num="pl" pdtype="pron" per="je" persoon="2v" pos="pron"
        postag="VNW(pers,pron,stan,nadr,2v,mv)" pt="vnw" rel="{rel}" rnum="pl" root="jullie" sense="jullie"
        status="nadr" vwtype="pers" wh="nwh" word="jullie"/>
"""

wetenwistnodetemplate = """
        <node begin="{begin}" end="{end}" frame="verb(hebben,past(sg),transitive)"  infl="sg" lcat="smain"
        lemma="weten" pos="verb" postag="WW(pv,verl,ev)" pt="ww" pvagr="ev" pvtijd="verl" rel="{rel}"
        root="weet" sc="transitive" sense="weet"  tense="past" word="wist" wvorm="pv"/>

"""

wetenwistennodetemplate = """
        <node begin="{begin}" end="{end}" frame="verb(hebben,past(pl),transitive)"  infl="pl" lcat="smain"
        lemma="weten" pos="verb" postag="WW(pv,verl,mv)" pt="ww" pvagr="mv" pvtijd="verl" rel="{rel}"
        root="weet" sc="transitive" sense="weet"  tense="past" word="wisten" wvorm="pv"/>

"""

zijnwasnodetemplate = """
        <node begin="{begin}" end="{end}" frame="verb(unacc,past(sg),copula)"  infl="sg" lcat="smain" lemma="zijn"
        pos="verb" postag="WW(pv,verl,ev)" pt="ww" pvagr="ev" pvtijd="verl" rel="{rel}" root="ben" sc="copula"
        sense="ben"  tense="past" word="was" wvorm="pv"/>

"""

zijnwarennodetemplate = """
        <node begin="{begin}" end="{end}" frame="verb(unacc,past(pl),copula)" infl="pl" lcat="smain" lemma="zijn"
        pos="verb" postag="WW(pv,verl,mv)" pt="ww" pvagr="mv" pvtijd="verl" rel="{rel}" root="ben" sc="copula"
        sense="ben"  tense="past" word="waren" wvorm="pv"/>

"""

pvnodestringtemplate = """
<node begin="{begin}" end="{end}"  lemma="{lemma}" pos="verb" postag="{postag}" pt="ww" pvagr="{pvagr}"
 pvtijd="{pvtijd}"  word="{word}" wvorm="{wvorm}"/>
"""

precedingjequerytemplate = ".//node[@word='je' and @end='{nounbegin}']"


def findfirst(nodelist, att, values):
    '''

    :param nodelist: list of nodes
    :param att: attribute
    :param values: valueset
    :return: the leftmost node for which the value of the attribute att is contained in valueset
    '''
    result = None
    for node in nodelist:
        nodeatt = getattval(node, att)
        if nodeatt in values:
            result = node
            break
    return result


def getnode(nodetemplate, sourcenode):
    sourcenodebegin = getattval(sourcenode, 'begin')
    sourcenodeend = getattval(sourcenode, 'end')
    sourcenoderel = getattval(sourcenode, 'rel')
    nodestring = nodetemplate.format(begin=sourcenodebegin, end=sourcenodeend, rel=sourcenoderel)
    result = etree.fromstring(nodestring)
    return result


def replaceverb(vnode):
    '''
    replaces nodes for less frequent verb forms by their more frequent counterparts
    :param vnode: node for a verb
    :return: a node for a verb
    '''
    vnodeword = getattval(vnode, 'word').lower()
    vnodelemma = getattval(vnode, 'lemma')
    if vnodeword == 'wist' and vnodelemma == 'wissen':
        result = getnode(wetenwistnodetemplate, vnode)
    elif vnodeword == 'wisten' and vnodelemma == 'wissen':
        result = getnode(wetenwistennodetemplate, vnode)
    elif vnodeword == 'was' and vnodelemma == 'wassen':
        result = getnode(zijnwasnodetemplate, vnode)
    elif vnodeword == 'waren' and vnodelemma == 'waren':
        result = getnode(zijnwarennodetemplate, vnode)
    else:
        result = vnode
    return result


def ptsubjcheck(child):
    results = []
    childpt = getattval(child, 'pt')
    childspecial = getattval(child, 'special')
    childword = getattval(child, 'word').lower()
    childlemma = getattval(child, 'lemma')
    childrel = getattval(child, 'rel')
    childbegin = getattval(child, 'begin')
    childend = getattval(child, 'end')

    if childlemma == 'het' and childrel in ['--']:
        results.append(child)
    elif childword == 'zij' and childpt == 'ww':
        zijsgnodestring = zijsgnodestringtemplate.format(begin=childbegin, end=childend, rel=childrel)
        zijsgnode = etree.fromstring(zijsgnodestring)
        results.append(zijsgnode)
    elif childpt in nominalpts and childspecial != "er_loc":
        results.append(child)
    elif childpt in nominalisablepts and getattval(child, 'positie') == 'nom':
        results.append(child)
    return results


def getpotsubjs(tree):
    results = []
    for child in tree:
        if 'pt' in child.attrib:
            childpt = getattval(child, 'pt')
            childcase = getattval(child, 'naamval')
            childvwtype = getattval(child, 'vwtype')
            childword = getattval(child, 'word').lower()
            if childword == 'zij':
                zijnode = getnode(zijsgnodestringtemplate, child)
                results.append(zijnode)
            #elif childword == 'ze':  # uitgezte want het levert fouten op
            #    zenode = getnode(zesgnodestringtemplate, child)
            #    results.append(zenode)
            elif childpt == 'vnw' and childvwtype == 'pers' and childcase == 'obl' and childword not in ['je']:
                pass  # exclude pronouns such as me, hem, mij, etc as subjects
            else:
                results += ptsubjcheck(child)
        elif 'cat' in child.attrib:
            childcat = getattval(child, 'cat')
            childrel = getattval(child, 'rel')
            if childcat == 'np':
                childhead = getheadof(child)
                if 'pt' in childhead.attrib:
                    childheadpt = getattval(childhead, 'pt')
                    childheadword = getattval(childhead, 'word')
                    if childheadpt == 'ww' or (childheadpt == 'n' and childheadword == 'waren'):
                        childdet = getdetof(child)
                        childdetbegin = getattval(child, 'begin')
                        childdetend = getattval(child, 'end')
                        childdetword = getattval(childdet, 'word').lower()
                        if childdetword == 'je':
                            jenode = getnode(jenodestringtemplate, childdet)
                            results.append(jenode)
                        elif childdetword == 'het':
                            hetnode = getnode(hetnodestringtemplate, childdet)
                            results.append(hetnode)
                        elif childdetword == "'t":
                            tnode = getnode(tnodestringtemplate, childdet)
                            results.append(tnode)
                    else:
                        childresults = ptsubjcheck(childhead)
                        results += childresults
                else:
                    childresults = getpotsubjs(child)  # here something should be done for coordinations
                    results += childresults
                #results.append(child)
            elif childcat in ['pp', 'inf', 'ppart'] and childrel not in ['dp', '--']:
                pass
            else:
                childresults = getpotsubjs(child)
                results += childresults
    return results


def getpvs(tokensmd: TokenListMD, tree: SynTree, uttid: UttId) -> List[SynTree]:
    """
    The function *getpvs* searches for finite verbs and potentially finite verbs in the structure

    * It first looks for words that are explicitly marked as finite verbs (*@pt="ww" and @wvorm="pv"*)
    except for the words *ja*, *zij* and *kijk*.
    * If none has been found, it identifies infinitrves as potential finite verbs (infinitives in Dutch are always
    homophonous to present tense plural finite verbs). However, such an infinitive should not be preceded by explicit
    infinitive markers such as *te* or *aan het* (as determined by the functions *ahi* and *ti*):

        * .. autofunction:: sva::ahi

        * .. autofunction:: sva::ti

        If multiple infinitives are found, arbitrarily the first one is selected.

        **Remark** We should actually select the first one that is not preceded by *te* or *aan het*.

        It the seelcted node is not immediately preceded by *te* or *aan het*
        @@hier verder@@
    """
    if tree is None:
        return []
    else:
        pvs = tree.xpath(".//node[@pt='ww' and @wvorm='pv' and @word!='zij' and @word !='kijk' and @word!='ja']")  # we exclude the conjunctive 'zij'
        # if we do not find a finite verb we analyse an infinitive as if it is a finite verb unless preceded by aan het or te
        if pvs == []:
            pvs = tree.xpath(".//node[@pt='ww' and @wvorm='inf']")
            if pvs != []:
                thepv = pvs[0]
                if not ahi(thepv, tree) and not ti(thepv, tree):
                    rb = rbrother(thepv, tree)    # use infl_rbrother instead for inflated trees
                    lb = lbrother(thepv, tree)    # use infl_lbrother instead for inflated trees
                    lblemma = getattval(lb, 'lemma')
                    lbrel = getattval(lb, 'rel')
                    if nominal(rb) or getattval(rb, 'pt') == 'lid' or (lblemma in ['je', 'het'] and lbrel == 'det'):
                        pass
                    else:
                        pvs = []
                else:
                    pvs = []
    # originally: but I do not understand it
    #            if nominal(rb):
    #                pvs = []
    #            if not(lblemma in ['je', 'het'] and lbrel == 'det'):
    #                pvs = []
            if debug and pvs != []:
                print('pv instead of infinitive')
                simpleshow(tree)

        # if we still did not find anything, try nouns that are also a verb form but restrict to cases
        # with possessive je present, or with verb zij present
        if pvs == []:
            nouns = tree.xpath(".//node[@pt='n' or @pt='adj']")
            zijverbs = tree.xpath(".//node[@pt='ww' and @word='zij']")
            for noun in nouns:
                nounword = getattval(noun, 'word')
                nounpt = getattval(noun, 'pt')
                nounstr = nounword.lower()
                nounbegin = getattval(noun, 'begin')
                nounend = getattval(noun, 'end')
                zijverbfound = zijverbs != []
                precedingjequery = precedingjequerytemplate.format(nounbegin=nounbegin)
                precedingjes = tree.xpath(precedingjequery)
                precedingjefound = precedingjes != []
                wistfound = nounstr == 'wist'
                if not (zijverbfound or precedingjefound or wistfound):
                    continue
                if informlexiconpos(nounstr, 'ww'):
                    cands = getwordposinfo(nounstr, 'ww')
                    if cands == []:
                        pvs = []
                    else:
                        if debug:
                            print('pv instead of noun or adj')
                            simpleshow(tree)
                        first = cands[0]
                        (pos, dehet, infl, lemma) = first
                        dcoi_infl = pvinfl2dcoi(nounstr, infl, lemma)
                        (wvorm, pvtijd, pvagr) = dcoi_infl
                        postag = 'WW({wvorm}, {pvtijd}, {pvagr})'.format(wvorm='pv', pvtijd=pvtijd, pvagr=pvagr)
                        pvnodestring = pvnodestringtemplate.format(lemma=lemma, word=nounstr, postag=postag, wvorm=wvorm,
                                                                   pvagr=pvagr, pvtijd=pvtijd, begin=nounbegin, end=nounend)
                        pvnode = etree.fromstring(pvnodestring)
                        pvs.append(pvnode)
        return pvs


def zijnimperativeok(vnode):
    vnodelemma = getattval(vnode, 'lemma')
    vnodeword = getattval(vnode, 'word').lower()
    vnodestype = getattval(vnode, 'stype')
    if vnodestype != 'imparative':
        result = True
    elif vnodelemma != 'zijn':
        result = True
    elif vnodeword in ['wees', 'weest']:
        result = True
    else:
        result = False
    return result


def ahi(node: SynTree, tree: SynTree) -> bool:
    '''
    The function *ahi* checks whether a given node *node* is immediately preceded by the words *aan* and *het* in
    syntactic structure *tree*.

    **Remark** The function currently works for standard syntactic structures, not for inflated
    syntactic structures.
    :param node: the node for which it is checked whether it is immediately preceded by *aan* and *het*
    :param tree: the syntactic structure that contains the node *node*.
    :return: True if *aan het* immediately precedes *node*, False otherwise
    '''
    nodepos = int(getattval(node, 'end'))
    p1end = str(nodepos - 2)
    p2end = str(nodepos - 1)
    p1 = find1(tree, '//node[@pt="vz" and @end={} and @lemma="aan"]'.format(p1end))
    p2 = find1(tree, '//node[@pt="lid" and @end={} and @lemma="het"]'.format(p2end))
    result = p1 is not None and p2 is not None
    return result


def ti(node: SynTree, tree: SynTree) -> bool:
    '''
    The function *ti* checks whether a given node *node* is immediately preceded by the word *te* in
    syntactic structure *tree*

    **Remark** The function currently works for standard syntactic structures, not for inflated
    syntactic structures.

    :param node: the node for which it is checked whether it is immediately preceded by *te*
    :param tree: the syntactic structure that contains the node *node*.
    :return: True if *te* immediately precedes *node*, False otherwise
    '''
    nodepos = int(getattval(node, 'end'))
    p1end = str(nodepos - 1)
    p1 = find1(tree, '//node[@pt and @end={} and @lemma="te"]'.format(p1end))
    result = p1 is not None
    return result


def getsvacorrectedutt(snode, thepv, tokens, metadata):
    newtokens = []
    pvbegin = getattval(thepv, 'begin')
    inversion = inverted(snode, thepv)
    reducedtokens = [t for t in tokens if not t.skip]
    newpv = getpvform(snode, thepv, inversion)
    if newpv is None:
        results = []
    else:
        newpos = int(pvbegin)
        newtoken = Token(newpv, newpos)
        for token in tokens:
            if token.pos != newpos:
                newtokens.append(token)
            else:
                oldtoken = token
                newtokens.append(newtoken)
                meta = mkSASTAMeta(oldtoken, newtoken, name='GrammarError', value='SVAerror', cat='Error',
                                   backplacement=bpl_node)
                metadata.append(meta)

        results = [TokenListMD(newtokens, metadata)]
    return results


def getsvacorrections(tokensmd: TokenListMD, rawtree: SynTree, uttid: UttId) -> List[TokenListMD]:
    '''

    :param tokensmd: the input sequence of tokens plus metadata
    :param rawtree: the syntacti structuire of the original utterance
    :param uttid: the identifier of the utterance
    :return: a list of pairs (token sequence , metadata)

    The function *getsvacorrections* generates alternative token sequences plus associated metadata for an input token
    sequence with associated metadata for an utterance with id *uttid* and (inflated) syntactic structure *rawtree*.

    The function performs the following steps:
    * It transforms all index nodes in the syntactic structure to nodes of their antecedents
    (except for the *rel* attribute) by means of the function *indextransform* from the module *treebankfunctons*. See  :ref:`indexnodeexpansion`.
    * It looks up all finite verb forms in the structure by means of the function *getpvs*
        * ..autofunction::  sva::getpvs

    * It only deals with structures that contain at most one finite verb
    * Next, it searches for the subject or for potential subjects in the structure. How it searches depend on the structure of the sentence.
        * normal sentences that match the  Xpath expression *normalsentencexpath*:

            .. autodata:: sva::normalsentencexpath

            In this case the query *subjxpath* is used to search for the subject(s).
            Of the subjects found (if any), the first one is selected as the subject.
            Normally, there is only one subject in a clause.

        * abnormalobj2 sentences that match the  Xpath expression *abnormalobj2sentencexpath*:

            .. autodata:: sva::abnormalobj2sentencexpath

            In this case, potential subjects are searched for, as described below.

        * normal whquestion sentences that match the  Xpath expression *normalwhqsentencexpath*:

            .. autodata:: sva::normalwhqsentencexpath

            In this case the query *subjxpath* is used to search for the subject(s).
            Of the subjects found (if any), the first one is selected as the subject.
            Normally, there is only one subject in a clause.

          The query *subjxpath* is defined as follows:

          .. autodata:: sva::subjxpath


    '''
    debug = False
    if debug:
        showtree(rawtree, text='rawtree')
    if rawtree is None:
        return []
    else:
        tree = indextransform(rawtree)
        tokens = tokensmd.tokens
        metadata = tokensmd.metadata
        ltokens = len(tokens)
        newtokens = []

        pvs = getpvs(tokensmd, tree, uttid)
        abnormalobj2matches = tree.xpath(abnormalobj2sentencexpath)
        if len(pvs) != 1:
            results = []
        elif (tree.xpath(normalsentencexpath) != [] and abnormalobj2matches == []) or \
                tree.xpath(normalwhqsentencexpath) != []:
            thesubjs = tree.xpath(subjxpath)
            if thesubjs != []:
                thesubj = thesubjs[0]
            else:
                thesubj = None
            prednode = getcopulapredicate(tree)
            if prednode is not None:
                thesubj = prednode
            thepv = pvs[0]
            if thesubj is not None:
                if phicompatible(thesubj, thepv):
                    results = []
                else:
                    results = getsvacorrectedutt(thesubj, thepv, tokens, metadata)
            else:
                results = []  # hier rekening houden met imperatieven!
        else:
            potsubjs = []
            rawpv = pvs[0]
            thepv = replaceverb(rawpv)
            pvstype = getattval(thepv, 'stype')
            if abnormalobj2matches != []:
                potsubjs = tree.xpath(abnormalobj2xpath)
            if potsubjs == []:
                potsubjs = getpotsubjs(tree)
            if potsubjs == []:
                results = []
            else:
                nominativenode = findfirst(potsubjs, 'naamval', {'nomin'})
                if nominativenode is not None:
                    thesubj = nominativenode
                # next left out because dealt with in a different way
                #   elif pvstype == 'imparative' and zijnimperativeok(thepv) and not modalinv(thepv):
                #       sunode = findfirst(potsubjs, 'rel', {'su'})
                #      if sunode is not None:
                #          thesubj = sunode
                #      else:
                #          sunode = findfirst(potsubjs, 'lemma', {'je', 'jij', 'jullie', 'u'})
                #          if sunode is not None:
                #              thesubj = sunode
                #          else:
                #              thepvend = getattval(thepv, 'end')
                #              jenodestring = jenodestringtemplate.format(begin=thepvend, end=thepvend, rel='su' )
                #              jenode = etree.fromstring(jenodestring)
                #              thesubj = jenode
                else:
                    sortedpotsubjs = sorted(potsubjs, key=lambda x: getattval(x, 'end'))
                    thesubj = sortedpotsubjs[0]
                thesubjlemma = getattval(thesubj, 'lemma')
                if thesubjlemma == 'u':
                    usgnode = getnode(usgnodestringtemplate, thesubj)
                    thesubj = usgnode
                thesubjpersoon = getattval(thesubj, 'persoon')
                if thesubjpersoon == 'persoon':
                    thesubj = copymodifynode(thesubj, {'persoon': '3'})
                if phicompatible(thesubj, thepv):
                    results = []
                else:
                    results = getsvacorrectedutt(thesubj, thepv, tokens, metadata)

        return results


def getpersons(vnode):
    pvagr = getattval(vnode, 'pvagr')
    vword = getattval(vnode, 'word')
    vtense = getattval(vnode, 'pvtijd')
    if vtense == 'tgw':
        if vword in {'ben', 'heb'}:
            result = {'1', '2i'}
        elif vword in {'bent', 'hebt'}:
            result = {'2', 'u'}
        elif vword in {'is', 'heeft'}:
            result = {'3', 'u'}
        elif vword in {'kun', 'zul'}:
            result = {'2i'}
        elif vword in {'kunt', 'zult', 'wilt'}:
            result = {'2', 'u'}
        elif vword in {'mag', 'kan', 'wil', 'zal'}:
            result = {'1', '2', '2i', '3', 'u'}
        elif pvagr == 'ev':
            if vword.endswith('t'):
                result = {'1', '2', '2i', '3', 'u'}
            else:
                result = {'1', '2i'}
        elif pvagr == 'met-t':
            result = {'2', '3', 'u'}
        elif pvagr == 'mv':
            result = {'mv'}
    elif vtense == 'verl':
        if pvagr == 'ev':
            result = {'1', '2', '2i', '3', 'u'}
        elif pvagr == 'mv':
            result = {'mv'}
        else:
            result = {}
    elif vtense == '':  # for infinitives
        result = {'mv'}
    elif vtense == 'conj':
        result = {'3'}
    else:
        settings.LOGGER.error('Unknown pvtijd value: {}'.format(vtense))
        result = {}
    return result


def phicompatible(snode, vnode):
    if snode is None or vnode is None:
        return False
    subjnode = snode
    subjnodelemma = getlemma(subjnode)
    inversion = inverted(subjnode, vnode)
    if subjnodelemma in ['het', 'u', 'je']:
        subjgetal = 'ev'
    else:
        subjgetal = getattval(subjnode, 'getal')
    vnodepvagr = getattval(vnode, 'pvagr')
    if subjgetal == 'mv':
        result = vnodepvagr == 'mv'
    elif subjgetal == 'getal':
        #check if the person corresponds with verbperson
        vnodepersons = getpersons(vnode)
        subjperson1 = getattval(subjnode, 'persoon')
        if subjperson1 == 'persoon':
            subjperson1 = '3'
        subjperson2 = subjperson1 if subjperson1 != '' else '3'
        subjperson = subjperson2[0]   # just the first character to deal with 2b 2v 3v 3p 30 etc
        if vnodepersons == {'mv'}:
            result = True
        elif subjperson in vnodepersons and not inversion:
            result = True
        else:
            result = False
    elif subjgetal == '':
        result = True
    elif subjgetal == 'ev':
        vnodepersons = getpersons(vnode)
        subjperson1 = getattval(subjnode, 'persoon')
        subjperson2 = subjperson1 if subjperson1 != '' else '3'
        subjperson = subjperson2[0]   # just the first character to deal with 2b 2v 3v 3p 30 etc
        if vnodepersons == {'mv'}:
            result = False
        elif subjperson in vnodepersons and not inversion:
            result = True
        elif '2i' in vnodepersons:
            subjbegin = getattval(subjnode, 'begin')
            vnodeend = getattval(vnode, 'end')
            result = subjperson == '2' and '2i' in vnodepersons and subjbegin >= vnodeend and \
                subjnodelemma in ['jij', 'je']
        elif 'u' in vnodepersons:
            subjnodelemma = getattval(subjnode, 'lemma')
            result = subjnodelemma == 'u'
        else:
            if subjperson not in ['1', '2', '3']:
                subjnodelemma = getattval(subjnode, 'lemma')
                settings.LOGGER.error('Unexpected value ({}) for the attribute persoon of {}'.format(subjperson, subjnodelemma))
            result = False
    else:
        subjnodelemma = getattval(subjnode, 'lemma')
        settings.LOGGER.error('Unexpected value ({}) for the attribute getal of {}'.format(subjgetal, subjnodelemma))
        result = False
    return result


def getpvform(thesubj, thepv, inversion):
    '''
    find in lexicon (CELEX)the verb wordform compatible with thesubj
    :param thesubj: subject (head) node
    :param thepv: finite verb
    :param inversion: True or False
    :return:
    '''

    results = getinflforms(thesubj, thepv, inversion)
    if results != []:
        result = results[0]
    else:
        result = None
    return result


predicatesubjxpath = """//node[@rel="predc" and
       parent::node[node[@rel="su" and (@lemma="het" or @lemma="dit" or @lemma="dat")] and
                    node[@rel="hd" and @pt="ww" and @wvorm="pv"]
                   ]
    ]"""


def getcopulapredicate(tree):
    preds = tree.xpath(predicatesubjxpath)
    if preds == []:
        result = None
    else:
        result = preds[0]
    return result



target_intarget, target_xsid, target_all, target_byrole, target_bysyn = 0, 1, 2, 3, 4
intargetxpath = '//meta[@name="intarget"]'
xsidxpath = '//meta[@name="xsid"]'
intargetvalxpath = './/meta[@name="intarget"]/@value'
xsidvalxpath = './/meta[@name="xsid"]/@value'
synxpath = './/meta[@name="syn"]'

rolevalxpath = './/meta[@name="role"]/@value'

targetroles = ['target_child', 'target', 'target_adult']


def get_targets(treebank):
    xsids = treebank.xpath(xsidxpath)
    intargets = treebank.xpath(intargetxpath)
    roles = treebank.xpath(rolevalxpath)
    targetrolesfound = any(map(lambda x: x.lower() in targetroles, roles))
    synannotations = treebank.xpath(synxpath)
    if synannotations != []:
        result = target_bysyn
    elif xsids != []:
        result = target_xsid
    elif intargets != []:
        result = target_intarget
    elif targetrolesfound:
        result = target_byrole
    else:
        result = target_all
    return result


def get_mustbedone(syntree, targets):
    if targets == target_bysyn:
        syns = syntree.xpath(synxpath)
        result = syns != []
    elif targets == target_intarget:
        intargetvals = syntree.xpath(intargetvalxpath)
        result = intargetvals != [] and intargetvals[0] == 'yes'
    elif targets == target_xsid:
        xsids = syntree.xpath(xsidvalxpath)
        result = xsids != []
    elif targets == target_byrole:
        rolevals = syntree.xpath(rolevalxpath)
        result = any(map(lambda x: x.lower() in targetroles, rolevals))
    else:
        result = True
    return result



from typing import Callable, List

from sastadev.macros import expandmacros
from sastadev.sastatypes import SynTree
from sastadev.treebankfunctions import (adjacent, find1, get_left_siblings,
                                        getattval, parent)

nietxpath = './/node[@lemma="niet"]'
wordxpath = './/node[@pt]'

vzn1basexpath = './/node[ @cat="pp" and (node[@pt="vz"] and node[(@pt="n" or @pt="vnw") and not (%Rpronoun%) and @rel="obj1"] and not(node[@pt="vz" and @vztype="fin"]))]'
vzn1xpath = expandmacros(vzn1basexpath)
vzn2xpath = './/node[node[@lemma="in" and @rel="mwp"] and node[@lemma="deze" and @rel="mwp"]]'
vzn3xpath = './/node[@pt="vz" and ../node[(@lemma="dit" or @lemma="dat")  and @begin>=../node[@pt="vz"]/@end and count(node)<=3] ]'
#vzn4basexpath = './/node[node[@pt="vz" and @rel="hd" and ../node[%Rpronoun% and @rel="obj1" and @end <= ../node[@rel="hd"]/@begin]]]'
#vzn4xpath = expandmacros(vzn4basexpath)

#: The constant *voslahbijxpath* selects nodes (PPs) that contain an adposition and an R-pronoun or a index node
#: coindexed with an R-pronoun.
#:
#: **Remark** It is not actually checked whether the indexed node has an R-pronoun as its antecedent
#:
#: **Remark** We may have to do something special for *pobj1*
#:
voslashbijxpath = expandmacros(""".//node[node[@pt="vz" and @rel="hd"] and
            node[@rel="obj1" and
                 ((@index and not(@word or @cat)) or
                  (%Rpronoun%)
                 )]]""")

#: The constant *vobijxpath* uses the macro *Vobij* to identify adverbial pronouns.
#: The macro **Vobij** is defined as follows::
#:
#:   Vobij = """(@pt="bw" and (contains(@frame,"er_adverb" ) or contains(@frame, "tmp_adverb") or @lemma="daarom") and
#:               @lemma!="er" and @lemma!="daar" and @lemma!="hier" and @lemma!="waar" and
#:               (starts-with(@lemma, 'er') or starts-with(@lemma, 'daar') or
#                 starts-with(@lemma, 'hier') or starts-with(@lemma, 'waar'))
#:              )"""
#:
vobijxpath = expandmacros('.//node[%Vobij%]')


def notadjacent(n1, n2, t): return not adjacent(n1, n2, t)


def xneg(stree):
    nodepairs = []
    nietnodes = stree.xpath(nietxpath)
    for nietnode in nietnodes:
        pnietnode = parent(nietnode)
        leftnietsiblings = get_left_siblings(nietnode)
        leftsiblings = get_left_siblings(pnietnode)
        ppnietnode = parent(pnietnode)
        if getattval(pnietnode, 'cat') == "advp" and len(leftsiblings) == 1 and getattval(ppnietnode, 'rel') == '--':
            result = True
            theleftsibling = leftsiblings[0]
        elif getattval(pnietnode, 'cat') != "advp" and getattval(pnietnode, 'rel') == '--' and len(leftnietsiblings) == 1:
            result = True
            theleftsibling = leftnietsiblings[0]
        else:
            result = False
        if result:
            nodepairs.append((theleftsibling, nietnode))
    if nodepairs == []:
        return None
    else:
        return nodepairs[0]


def xneg_neg(stree):
    (x, neg) = xneg(stree)
    return neg


def xneg_x(stree):
    (x, neg) = xneg(stree)
    return x


def VzN(stree):
    results = []
    results += stree.xpath(vzn1xpath)
    results += stree.xpath(vzn2xpath)
    results += stree.xpath(vzn3xpath)
    #results += stree.xpath(vzn4xpath) # does not belong here after all, these will be scored under Vo/Bij
    return results


def auxvobij(stree: SynTree, pred: Callable[[SynTree, SynTree, SynTree], bool]) -> List[SynTree]:
    '''

    :param stree: the syntactic structure to be analysed
    :param pred: a predicate that the results found must satisfy
    :return: a list of matching nodes

    The function *auxvobij* finds nodes that are found by the *voslashbijxpath* and selects from these those
    that satisfy the predicate *pred*. It is used to distinguish cases of R-pronoun + adposition that are *adjacent*
    (which should be analysed as TARSP *Vobij*) from those that are not adjacent (which should be analysed as TARSP
    Vo/Bij).

    .. autodata:: queryfunctions::voslashbijxpath

    '''
    RPnodes = stree.xpath(voslashbijxpath)
    results = []
    for RPnode in RPnodes:
        # find the head node
        headnode = find1(RPnode, 'node[@rel="hd"]')

        # find the obj1node
        obj1node = find1(RPnode, 'node[@rel="obj1"]')

        # check if they are adjacent
        if headnode is not None and obj1node is not None:
            if pred(headnode, obj1node, stree):
                results.append(RPnode)
    return results


def vobij(stree: SynTree) -> List[SynTree]:
    '''

    :param stree: syntactic structure to be analysed
    :return: List of matching nodes

    The function *vobij* uses the Xpath expression *vobijxpath* and the function *auxvobij* to obtain its resulting nodes:

    * The *vobijxpath* expression matches with so-called adverbial pronouns:

      .. autodata:: queryfunctions::vobijxpath

    * The function *auxvobij*  finds adjacent R-pronoun + adposition cases:

      .. autofunction:: queryfunctions::auxvobij

    '''
    results1 = stree.xpath(vobijxpath)
    results2 = auxvobij(stree, adjacent)
    results = results1 + results2
    return results


def voslashbij(stree: SynTree) -> List[SynTree]:
    '''

    :param stree: syntactic structuire to be analysed
    :return: List of matching nodes

    The function *voslashbij* uses the function *auxvobij* to find non-adjacent R-pronoun + adposition cases:

    .. autofunction:: queryfunctions::auxvobij
          :noindex:


    '''
    results = auxvobij(stree, notadjacent)
    return results



'''
The module *alpino* provides functions related to Alpino structures. In particular,
it provides the function *getdehetwordinfo*:

* obtaining grammatical information from nouns in an Alpino parse

  .. autofunction:: alpino::getdehetwordinfo

'''
#from __future__ import annotations

from typing import List, Tuple

from sastadev import lexicon, treebankfunctions
from sastadev.conf import settings
from sastadev.sastatypes import WordInfo


def getalpinowordinfo(word: str) -> List[WordInfo]:
    '''
    The function *getalpinowordinfo* parses the input word in isolation, extracts its properties and returns some of
    the properties as a list of WordInfo objects.

    Note that it is essential to parse the word in isolation, because unknown words
    that follow a uter determiner are analysed by Alpino as a uter noun
    (e.g. in *de ponyautootje*), even if they are analysed as neuter nouns in
    isolation.

    '''
    tree = settings.PARSE_FUNC(word)
    relevantnode = treebankfunctions.find1(tree, '//node[parent::node[@cat="top"]]')
    if relevantnode is None:
        return []
    else:
        pos = treebankfunctions.getattval(relevantnode, 'pt')
        if pos == 'n':
            genus = treebankfunctions.getattval(relevantnode, 'genus')
            dehet = lexicon.het if genus == 'onz' else lexicon.de
            getal = treebankfunctions.getattval(relevantnode, 'getal')
            graad = treebankfunctions.getattval(relevantnode, 'graad')
            infl = 'd' if graad == 'dim' else ''
            infl += 'e' if getal == 'ev' else 'm'
            lemma = treebankfunctions.getattval(relevantnode, 'lemma')
            return ([(pos, dehet, infl, lemma)])
        else:
            return []


def getdehetwordinfo(wrd: str) -> Tuple[List[WordInfo], str]:
    '''
    The function *getdehetwordinfo*  determines the properties of the input string *word* by first looking in the
    lexicon. It only includes properties if the word is a noun.
    If not found in the lexicon, it tries to find the properties from Alpino by means of the function
    *getalpinowordinfo*.
    It returns a tuple consisting of a list of WordInfo objects and a string indicating the source where the properties
    have been found.

    .. autofunction:: alpino::getalpinowordinfo
    '''

    wordinfos = lexicon.getwordinfo(wrd)

    # we only want to consider nouns or words of unknown word class (such as kopje in CELEX)
    wordinfos = [wordinfo for wordinfo in wordinfos if wordinfo[0] in ['n', 'None']]
    # if any of the alternatives is a de-word, we keep only these
    dewordinfos = [wordinfo for wordinfo in wordinfos if wordinfo[1] == lexicon.de]
    if dewordinfos != []:
        wordinfos = dewordinfos
    #if any([wordinfo[1] == lexicon.de for wordinfo in wordinfos]):
    #    wordinfos = []

    # if not found yet we check with Alpino
    if wordinfos != []:
        source = 'celex'
    else:
        wordinfos = getalpinowordinfo(wrd)
        source = 'alpino'
    return (wordinfos, source)



import re

from sastadev import cleanCHILDEStokens
from sastadev.conf import settings
from sastadev.metadata import Meta, bpl_delete
from sastadev.sastatoken import Token, show

CHAT = 'CHAT'

CHAT_explanation = 'Explanation'

monadic = 1
dyadic = 2

emptyregex = r''
eps = ''
space = ' '
keep = r'\0'
restart = 'restart'
scope_open = '<'
scope_close = '>'
emptyreplacement = eps
anybutrb = r'[^\]]*'

errormarking = 'Error Marking'
omittedword = 'Omitted Word'
specialform = 'Special Form'


def fullre(pat):
    result = r'^' + pat + r'$'
    return result


def identity(x):
    return x


def refunction(x):
    '''
    The input regular expression is adapted to the requirements, surrounded by ^ and $ (see refunction)
    :param x: a regular expression
    :return: a modified regular expression
    '''
    result = fullre(x)
    return result


# u2013 = en-dash, u2014 = em-dash, u2015 = horizontal bar


# states
bstate, mstate, estate = 0, 1, 2
scopestate, scopefoundstate, wstate, rfoundstate = 11, 12, 13, 14
wordinitial = r'[\w\'\(\)' + u'\u2019\u02C8\u02CC\u02de\u21AB\u2260' + r']'
wordpat = wordinitial + r'[\w:\^\+_~\'\-\(\)' + u'\u2019\u02C8\u02CC\u0329\u02de\u21AB\u2260' + r']*'  # removed \. here in the second part maybe needed for (.)?
simplewordpat = r'\w+'
fullwordpat = fullre(wordpat)
wordre = re.compile(fullwordpat)
# interpunction = r'(:?' + r'[!\?\.,;]' + '|' + u'[\u201C\u201D\u2039\u203A]' + r'|' + r'(?<=\s):' + r')'
interpunction = r'\-\-\-|\-\-|\-|\-' + r'|' + r'[!\?\.,;]' + '|' + u'[\u2013\u2014\u2015\u201C\u201D\u2039\u203A]' + r'|' + r'(?<=\s):'
filenamepat = r'[\w\.]+'
fullfilenamepat = fullre(filenamepat)
fullfilenamere = re.compile(fullfilenamepat)

wordorpuncpat = r'(:?' + wordpat + '|' + interpunction + ')'

specialformpat = wordpat + r'(?:@z:\w\w\w|@\w\w?\w?)'
fullspecialformpat = fullre(specialformpat)
specialformre = re.compile(fullspecialformpat)
repkeepannotations = ['Repetition', 'Retracing', 'Reformulation']


def getreplacement(repkeep, annotation):
    if annotation.name in repkeepannotations:
        if repkeep:
            replacement = keep
        else:
            replacement = eps
    else:
        replacement = annotation.regex.replacement
    return replacement


def doreplacement(repltokens, replacement, tokens):
    newtokens = tokens
    if replacement == keep:
        newtokens += repltokens
    elif replacement == eps:
        pass
    else:
        settings.LOGGER.error('Unknown replacement: {}'.format(replacement))
    return newtokens


class CHAT_Annotation:
    def __init__(self, name, v2015, v2020, regex, metadatafunction, message=None):
        self.name = name
        self.v2015 = v2015
        self.v2020 = v2020
        self.regex = regex
        self.metadatafunction = metadatafunction
        self.message = message

    def apply(self, tokens, repkeep=False):
        (newtokens, metadata) = self.regex.apply(tokens, self, repkeep)
        if self.message is not None:
            self.message(self.name, show(tokens))
        return (newtokens, metadata)


def isaword(wrd):
    match = wordre.search(wrd)
    result = match is not None
    return result


class CHAT_Regex:
    def __init__(self):
        pass


class CHAT_InWordRegex(CHAT_Regex):
    def __init__(self, regex, replacement):
        self.regex = regex
        self.replacement = replacement
        self.compiledre = re.compile(identity(self.regex))

    def apply(self, tokens, annotation, repkeep):
        newtokens = []
        metadata = []
        insidebrackets, outsidebrackets = 0, 1
        state = outsidebrackets
        for token in tokens:
            if isaword(token.word):
                if state == outsidebrackets:
                    matches = self.compiledre.finditer(token.word)
                    for m in matches:
                        annotationwordlist = [m.group()]
                        annotatedposlist = [token.pos]
                        annotatedwordlist = [token.word]
                        annotationposlist = [p for p in range(m.start(), m.end())]
                        newmeta = annotation.metadatafunction(annotation, annotationwordlist, annotatedposlist,
                                                              annotatedwordlist, annotationposlist)
                        metadata.append(newmeta)
                    newword = self.compiledre.sub(self.replacement, token.word)
                    newtoken = Token(newword, token.pos)
                else:
                    newtoken = token
            elif isopenbracket(token.word):
                state = insidebrackets
                newtoken = token
            elif isclosebracket(token.word):
                state = outsidebrackets
                newtoken = token
            else:
                newtoken = token
            if newtoken.word != "":
                newtokens.append(newtoken)
        return (newtokens, metadata)


class CHAT_SimpleRegex(CHAT_Regex):
    def __init__(self, regex, replacement, scoped):
        self.regex = regex
        self.replacement = replacement
        self.scoped = scoped
        self.compiledre = re.compile(refunction(self.regex))

    def apply(self, tokens, annotation, repkeep):
        newtokens = []
        metadata = []
        insidebrackets, outsidebrackets = 0, 1
        state = outsidebrackets
        for token in tokens:
            if self.compiledre.search(token.word):
                if state == outsidebrackets:
                    if self.replacement == keep:
                        newtoken = token
                        newtokens.append(newtoken)
                    elif callable(self.replacement):
                        newtokenword = self.replacement(token.word)
                        newtoken = Token(newtokenword, token.pos)
                        newtokens.append(newtoken)
                    elif isinstance(self.replacement, str):
                        newtokenword = self.compiledre.sub(self.replacement, token.word)
                        if newtokenword != '':
                            newtoken = Token(newtokenword, token.pos)
                            newtokens.append(newtoken)
                        else:
                            pass  # token is removed
                    else:
                        pass  # token is removed
                    metadata.append(annotation.metadatafunction(annotation, token.pos, token.word))
                else:
                    newtokens.append(token)
            elif isopenbracket(token.word):
                state = insidebrackets
                newtokens.append(token)
            elif isclosebracket(token.word):
                state = outsidebrackets
                newtokens.append(token)
            else:
                newtokens.append(token)

        return (newtokens, metadata)


class CHAT_SimpleScopedRegex(CHAT_Regex):
    def __init__(self, regex, replacement, scoped, arity):
        self.regex = regex
        self.replacement = replacement
        self.scoped = scoped
        self.arity = arity
        self.compiledre = re.compile(refunction(self.regex))

    def apply(self, tokens, annotation, repkeep):
        newtokens = []
        metadata = []
        todotokens = tokens
        ltodotokens = len(todotokens)
        tokenctr = 0

        # while todotokens != []:
        while tokenctr < ltodotokens:
            scope = findscope(todotokens[tokenctr:], offset=tokenctr)  # this finds a scope part surrounded by <>
            if scope is None:
                # hier misschien iets voor [///] en restart
                newtokens += todotokens[tokenctr:]
                # result = (newtokens, metadata)
                break
            else:
                (b, e) = scope
                if ltodotokens == e + 1:
                    settings.LOGGER.error(
                        'Scope markings in positions {} and {} not followed by annotation ignored in {}'.format(b, e,
                                                                                                                show(
                                                                                                                    todotokens)))
                    newtokens += todotokens[:b] + todotokens[b + 1:e]
                    tokenctr = e + 1
                elif self.compiledre.search(todotokens[e + 1].word):
                    annotationwords = [token.word for token in todotokens[b + 1:e]]
                    annotationpositions = [token.pos for token in todotokens[b + 1:e]]
                    if self.arity == dyadic:
                        if ltodotokens <= e + 2:
                            settings.LOGGER.error(
                                'Missing second argument for dyadic annotation {} in {}'.format(annotation.name,
                                                                                                show(todotokens)))
                            newtokens += todotokens[b + 1:e]
                            break
                        else:
                            annotatedwords = [todotokens[e + 2].word] if self.replacement == eps else []
                            annotatedpositions = [todotokens[e + 2].pos] if self.replacement == eps else []
                    elif self.arity == monadic:
                        annotatedwords = []
                        annotatedpositions = []
                    else:
                        settings.LOGGER.error('Illegal arity specification ({}) on {}'.format(self.arity, annotation.name))
                        annotatedwords = []
                        annotatedpositions = []
                    newmeta = annotation.metadatafunction(annotation, annotationwords, annotatedpositions,
                                                          annotatedwords, annotationpositions)
                    metadata.append(newmeta)
                    newtokens += todotokens[tokenctr:b]
                    replacement = getreplacement(repkeep, annotation)
                    newtokens = doreplacement(todotokens[b + 1:e], replacement, newtokens)
                    tokenctr = e + 2
                else:
                    newtokens += todotokens[tokenctr:e + 1]
                    # todotokens = todotokens[e+1:]
                    tokenctr = e + 1

        # the code with a scoped token sequence
        # look for the code
        todotokens = newtokens
        ltodotokens = len(todotokens)
        newtokens = []
        prevtoken = None
        scopewords = []
        scopepositions = []
        i = 0
        while i < ltodotokens:
            if self.compiledre.search(todotokens[i].word):
                if scopewords == []:
                    settings.LOGGER.error(
                        'First argument of annotation {} missing. Annotation ignored'.format(annotation.name))
                else:
                    if self.arity == monadic:
                        annotatedpositions = []
                        annotatedwords = []
                        replacement = getreplacement(repkeep, annotation)
                        newtokens = doreplacement([prevtoken], replacement, newtokens)
                        prevtoken = None
                        newmeta = annotation.metadatafunction(annotation, scopewords, annotatedpositions,
                                                              annotatedwords, scopepositions)
                        metadata.append(newmeta)
                    elif self.arity == dyadic:
                        if i + 1 >= ltodotokens:
                            settings.LOGGER.error(
                                'Missing second argument for dyadic annotation {} in {}'.format(annotation.name,
                                                                                                show(todotokens)))
                        else:
                            annotatedpositions = [todotokens[i + 1].pos]
                            annotatedwords = [todotokens[i + 1].word]
                            replacement = getreplacement(repkeep, annotation)
                            newtokens = doreplacement([prevtoken], replacement, newtokens)
                            prevtoken = None
                            newmeta = annotation.metadatafunction(annotation, scopewords, annotatedpositions,
                                                                  annotatedwords, scopepositions)
                            metadata.append(newmeta)
            else:
                if prevtoken is not None:
                    newtokens.append(prevtoken)
                scopewords = [todotokens[i].word]
                scopepositions = [todotokens[i].pos]
                prevtoken = todotokens[i]

            i += 1
        if prevtoken is not None:
            newtokens.append(prevtoken)
        return (newtokens, metadata)


class CHAT_ComplexRegex(CHAT_Regex):
    def __init__(self, regextuple, replacementtuple, scoped, containswords=False):
        self.regexbegin = regextuple[0]  # 3 elements: begin mid end
        self.regexmid = regextuple[1]  # 3 elements: begin mid end
        self.regexend = regextuple[2]  # 3 elements: begin mid end
        self.scopereplacement = replacementtuple[0]  # 2 elements: one for the scope and one for the text between [ ]
        self.bracketreplacement = replacementtuple[1]  # 2 elements: one for the scope and one for the text between [ ]
        self.scoped = scoped
        self.containswords = containswords
        self.compiledrebegin = re.compile(refunction(self.regexbegin))
        self.compiledremid = re.compile(refunction(self.regexmid))
        self.compiledreend = re.compile(refunction(self.regexend))

    def apply(self, tokens, annotation, repkeep):
        bracketregexes = (self.compiledrebegin, self.compiledremid, self.compiledreend)
        metadata = []
        estates = [bstate, wstate]
        state = bstate
        tokenctr = 0
        todotokens = tokens
        newtokens = []
        tobereplacedtokens = []
        while tokenctr < len(tokens):
            token = todotokens[tokenctr]
            inc = 1
            if state == bstate:
                if token.word == '<':
                    state = scopestate
                elif wordre.search(token.word):
                    state = wstate
                    tobereplacedtokens = [token]
                elif self.compiledrebegin.search(token.word):
                    state = rfoundstate
                else:
                    state = bstate
                    newtokens.append(token)
            elif state == wstate:
                if token.word == '<':
                    newtokens += tobereplacedtokens
                    tobereplacedtokens = []
                    state = scopestate
                elif self.compiledrebegin.search(token.word):
                    state = rfoundstate
                elif wordre.search(token.word):
                    state = wstate
                    newtokens += tobereplacedtokens
                    tobereplacedtokens = [token]
                else:
                    state = wstate
                    newtokens += tobereplacedtokens
                    tobereplacedtokens = [token]
            elif state == scopestate:
                scope = findscope(tokens[tokenctr - 1:], offset=tokenctr - 1)
                if scope is None:
                    settings.LOGGER.error('No closing bracket found for < with pos={} in {}'.format(
                        tokens[tokenctr - 1].pos, show(tokens)))
                    state = wstate
                else:
                    (b, e) = scope
                    tobereplacedtokens = todotokens[b:e + 1]
                    state = wstate
                    inc = e - tokenctr + 1
            elif state == rfoundstate:
                bbbe = findbrackets(todotokens[tokenctr - 1:], bracketregexes, offset=tokenctr - 1)
                if bbbe is not None:
                    (bracketbegin, bracketend) = bbbe
                    annotationtokens = todotokens[bracketbegin + 1: bracketend]
                    (cleanannotationtokens, innermetadata) = cleanCHILDEStokens.cleantokens(annotationtokens,
                                                                                            repkeep) if self.containswords else (
                        annotationtokens, [])
                    metadata += innermetadata
                    annotatedwords = [t.word for t in tobereplacedtokens if t.word not in ['<', '>']]
                    annotatedpositions = [t.pos for t in tobereplacedtokens if t.word not in ['<', '>']]
                    thevalue = [token.word for token in cleanannotationtokens]
                    annotationpositions = [token.pos for token in cleanannotationtokens]
                    newmeta = annotation.metadatafunction(annotation, thevalue, annotatedpositions, annotatedwords,
                                                          annotationpositions)
                    metadata.append(newmeta)
                    replacement = self.scopereplacement
                    repltokens = [t for t in tobereplacedtokens if t.word not in ['<', '>']]
                    newtokens = doreplacement(repltokens, replacement, newtokens)
                    if self.bracketreplacement == keep:
                        newtokens += cleanannotationtokens
                    elif self.bracketreplacement == eps:
                        pass
                    else:
                        settings.LOGGER.error('Unknown replacementtype: {} in {}'.format(self.scopereplacement, show(tokens)))
                    tobereplacedtokens = []
                    inc = bracketend - bracketbegin
                state = wstate
            tokenctr += inc
        newtokens += tobereplacedtokens
        if state in estates:
            return (newtokens, metadata)
        else:
            settings.LOGGER.error('Not in an end state, state={} in {}'.format(state, show(tokens)))
            return (tokens, [])


def findbrackets(tokens, regexes, offset=0):
    (bregex, mregex, eregex) = regexes
    b1e1 = bracketseq(tokens, bregex, mregex, eregex)
    if b1e1 is not None:
        (b1, e1) = b1e1
        result = (b1 + offset, e1 + offset)
    else:
        result = None
    return result


def getsfvalue(w):
    index = w.find('@')
    result = w[index:]
    return result


def getsfword(w):
    index = w.find('@')
    result = w[:index]
    return result


def dropbrackets(w):
    result1 = [c for c in w if c not in ['(', ')']]
    result = eps.join(result1)
    return result


def simplemetafunction(f):
    return lambda ann, pos, w: Meta(ann.name, [f(w)],
                                    annotatedposlist=[pos],
                                    annotatedwordlist=[w],
                                    source=CHAT)


def simple_bpldel_metafunction(f):
    return lambda ann, pos, w: Meta(ann.name, [f(w)],
                                    annotatedposlist=[pos],
                                    annotatedwordlist=[w], source=CHAT,
                                    backplacement=bpl_delete)


def simplescopedmetafunction(ann, annotationwordlist, annotatedposlist, annotatedwordlist, annotationposlist):
    return Meta(ann.name, annotationwordlist,
                annotationposlist=annotationposlist, annotatedposlist=annotatedposlist,
                annotatedwordlist=annotatedwordlist, source=CHAT)


def complexmetafunction(ann, annotationwordlist, annotatedposlist, annotatedwordlist, annotationposlist):
    return Meta(ann.name, annotationwordlist,
                annotationposlist=annotationposlist, annotatedwordlist=annotatedwordlist,
                annotatedposlist=annotatedposlist, source=CHAT)


def charmetafunction(ann, annotationcharlist, annotatedcharlist, annotationcharposlist, annotatedcharposlist):
    return Meta(ann.name, annotationcharlist, annotationcharlist=annotationcharlist,
                annotatedcharlist=annotatedcharlist,
                annotationcharposlist=annotationcharposlist, annotatedcharposlist=annotatedcharposlist)


def epsf(w):
    return ''


interposedpat = r'^&\*(\w\w\w:[\w:]+)$'
interposedre = re.compile(interposedpat)


def interposedword(w):
    result1 = interposedre.findall(w)
    result = ''.join(result1)
    return result


def dropinitial(w):
    if w is None:
        result = None
    else:
        result = w[1:]
    return result


def dropzero(w):
    if w is None:
        result = None
    elif w[0] == '0':
        result = w[1:]
    else:
        result = w
    return result


def dropsubstr(w, s):
    if w is None or s is None:
        result = w
    else:
        ls = len(s)
        sindex = w.find(s)
        result = w[:sindex] + w[sindex + ls:]
    return result


def dropchars(c):
    return lambda w: dropchars2(w, c)


def dropchars2(w, c):
    result1 = [ch for ch in w if ch != c]
    result = ''.join(result1)
    return result


def CHAT_message(msg):
    def result(x, y):
        return settings.LOGGER.warning(msg.format(x, y))

    return result


annotations = [
    # overlap crucially before annotations taht delete things
    CHAT_Annotation('Overlap Follows', '8.4:71', '10.3:74-75',
                    CHAT_SimpleScopedRegex(r'\[\>[0-9]?\]', keep, True, monadic), simplescopedmetafunction),
    # here additional things could be done
    CHAT_Annotation('Overlap Precedes', '8.4:71-72', '10.3:75',
                    CHAT_SimpleScopedRegex(r'\[\<[0-9]?\]', keep, True, monadic), simplescopedmetafunction),
    CHAT_Annotation(specialform, '6.3:37', '8.3:43-44', CHAT_SimpleRegex(specialformpat, getsfword, False),
                    simplemetafunction(getsfvalue)),
    CHAT_Annotation('Unintelligible Speech', '6.4:41', '8.4:47', CHAT_SimpleRegex(r'xxx', keep, False),
                    simplemetafunction(epsf)),
    CHAT_Annotation('Phonological Coding', '6.4:41', '8.4:47', CHAT_SimpleRegex(r'yyy', keep, False),
                    simplemetafunction(epsf)),
    CHAT_Annotation('Noncompletion of a Word', '6.5:43', '8.5:48',
                    CHAT_InWordRegex(r'\(([-\w\']*)\)', r'\1'), charmetafunction),
    CHAT_Annotation(omittedword, '6.5:43', '8.5:48-49',
                    CHAT_SimpleRegex(r'0[\w:]+', dropzero, False), simple_bpldel_metafunction(dropzero)),
    CHAT_Annotation('Satellite at End', '7.4:58', '9.2:59-60',
                    CHAT_SimpleRegex(r'\s„\s', eps, False), simplemetafunction(identity)),
    CHAT_Annotation('Satellite in Beginning', '7.4:58', '9.2:59-60',
                    CHAT_SimpleRegex(r'\s‡\s', eps, False), simplemetafunction(identity)),
    CHAT_Annotation('Falling Tone', '7.6:59', '9.8:63',
                    CHAT_SimpleRegex(u'\u2193', eps, False), simplemetafunction(identity)),
    CHAT_Annotation('Rising Tone', '7.6:59', '9.8:63',
                    CHAT_SimpleRegex(u'\u2191', eps, False), simplemetafunction(identity)),
    CHAT_Annotation('Primary Stress', '7.7:59', '9.9:63', CHAT_InWordRegex(u'\u02C8', ''), complexmetafunction),
    CHAT_Annotation('Secondary Stress', '7.7:59', '9.9:63', CHAT_InWordRegex(u'\u02CC', ''), complexmetafunction),
    CHAT_Annotation('Lengthened Syllable', '7.7:59-60', '9.9:63', CHAT_InWordRegex(r':', ''), complexmetafunction),
    CHAT_Annotation('Blocking', '7.7:60', '9.9:64', CHAT_SimpleRegex(r'\^' + wordpat, dropinitial, False),
                    simplemetafunction(dropinitial)),  # this one must crucially precede Pause Between Syllables
    CHAT_Annotation('Pause Between Syllables', '7.7:60', '9.9:63-64', CHAT_InWordRegex(r'\^', ''), complexmetafunction),
    CHAT_Annotation('Simple Event', '7.8.1:60', '9.10.1:64-65', CHAT_SimpleRegex(r'&=[\w:]+', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Complex Local Event', '7.8.2:61', '9.10.3:65',
                    CHAT_ComplexRegex((r'\[\^\s', wordorpuncpat, r'\]'), (keep, eps), False),
                    complexmetafunction),
    CHAT_Annotation('Pause', '7.8.3:62', '9.10.4:66', CHAT_SimpleRegex(r'\(\.\.?\.?\)', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Timed Pause', '7.8.3:62', '9.10.4:66', CHAT_SimpleRegex(r'\([0-9]+\.[0-9]+\)', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Long Event', '7.8.4:62', '9.10.5:66',
                    CHAT_ComplexRegex((r'&{l=[\w:]+', wordpat, r'&}l=[\w:]+'), (keep, eps), True),
                    complexmetafunction),  # no check that the latter part equals the initial part
    CHAT_Annotation('Long Nonverbal Event', '7.8.4:62', '9.10.5:66',
                    CHAT_ComplexRegex((r'&{n=[\w:]+', wordpat, r'&}n=[\w:]+'), (keep, eps), True),
                    complexmetafunction),  # no check that the latter part equals the initial part
    CHAT_Annotation('Trailing Off', '7.9:62', '9.11:66', CHAT_SimpleRegex(r'\+\.\.\.', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Trailing Off of a Question', '7.9:63', '9.11:67', CHAT_SimpleRegex(r'\+\.\.\?', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Question With Exclamation', '7.9:63', '9.11:67', CHAT_SimpleRegex(r'\+!\?', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Interruption', '7.9:63', '9.11:67', CHAT_SimpleRegex(r'\+/\.', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Interruption of a Question', '7.9:63', '9.11:67', CHAT_SimpleRegex(r'\+/\?', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Self-Interruption', '7.9:64', '9.11:68', CHAT_SimpleRegex(r'\+//\.', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Self-Interrupted Question', '7.9:64', '9.11:68', CHAT_SimpleRegex(r'\+//\?', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Transcription Break', '7.9:64', '9.11:68', CHAT_SimpleRegex(r'\+\.', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Quotation Begin', '7.9:64', '9.11:68', CHAT_SimpleRegex(u'\u201C', keep, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Quotation End', '7.9:64', '9.11:68', CHAT_SimpleRegex(u'\u201D', keep, False),
                    simplemetafunction(identity)),
    # CHAT_Annotation('Inner Quotation Begin', '7.9', '64', CHAT_SimpleRegex(u'\u2018', keep, False),
    #                simplemetafunction(identity)), # not anymore in the 2020-06-19 edition
    # CHAT_Annotation('Inner Quotation End', '7.9', '64', CHAT_SimpleRegex(u'\u2019', keep, False),
    #                simplemetafunction(identity)), # not anymore in the 2020-06-19 edition
    CHAT_Annotation('Quotation Follows', '7.9:64-65', '9.11:68-69', CHAT_SimpleRegex(r'\+\"/\.', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Quotation Precedes', '7.9:65', '9.11:69', CHAT_SimpleRegex(r'\+\"\.', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Quoted Utterance', '7.10:65', '9.11:69', CHAT_SimpleRegex(r'\+\"', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Quick Uptake', '7.10:65', '9.11:69', CHAT_SimpleRegex(r'\+\^', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Lazy Overlap', '7.10:65-66', '10.3:75', CHAT_SimpleRegex(r'\+\<', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Self Completion', '7.10:66', '9.11:69', CHAT_SimpleRegex(r'\+,', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Other Completion', '7.10:66', '9.11:69-70', CHAT_SimpleRegex(r'\+\+', eps, False),
                    simplemetafunction(identity)),

    # erroR marking crucially before [/] [//] [///] etc
    CHAT_Annotation(errormarking, '8.5:75', '10.5:78', CHAT_SimpleScopedRegex(r'\[\*\]', keep, True, monadic),
                    simplescopedmetafunction),
    CHAT_Annotation(errormarking, '8.5:75', '10.5:78',
                    CHAT_ComplexRegex((r'\[\*', r'[\w:\-\+=]+', r'\]'), (keep, eps), False),
                    complexmetafunction),

    CHAT_Annotation('Pic Bullet', '8.1:67', '10.1:71',
                    CHAT_ComplexRegex((u'\u00b7' + r'%pic:', filenamepat, u'\u00b7'), (keep, eps), True),
                    complexmetafunction),  # pic bullet and text bullet must essentially before time alignment
    CHAT_Annotation('Text Bullet', '8.1:67', '10.1:71',
                    CHAT_ComplexRegex((u'\u00b7' + r'%txt:', filenamepat, u'\u00b7'), (keep, eps), True),
                    complexmetafunction),
    CHAT_Annotation('Time Alignment', '7.10:67', '10.1:71',
                    CHAT_ComplexRegex((u'\u00b7', r'[0-9_]+', u'\u00b7'), (keep, eps), True),
                    complexmetafunction),
    CHAT_Annotation('Time Alignment', '7.10:67', '10.1:71',
                    CHAT_ComplexRegex((u'\u0015', r'[0-9_]+', u'\u0015'), (keep, eps), True),
                    complexmetafunction),  # not an official code but it occurs as such in CLPF
    CHAT_Annotation('Paralinguistic Material', '8.2:68', '10.1:72',
                    CHAT_ComplexRegex((r'\[=!', anybutrb, r'\]'), (keep, eps), True),
                    complexmetafunction),
    CHAT_Annotation('Stressing', '8.2:68', '10.1:72', CHAT_SimpleScopedRegex(r'\[!\]', keep, False, monadic),
                    simplescopedmetafunction),
    CHAT_Annotation('Contrastive Stressing', '8.2:68', '10.1:72',
                    CHAT_SimpleScopedRegex(r'\[!!\]', keep, False, monadic),
                    simplescopedmetafunction),
    # Duration to be added here @@
    CHAT_Annotation(CHAT_explanation, '8.3:69', '10.3:73',
                    CHAT_ComplexRegex((r'\[=', anybutrb, r'\]'), (keep, eps), False),
                    complexmetafunction),
    CHAT_Annotation('Replacement', '8.3:69', '10.3:73',
                    CHAT_ComplexRegex((r'\[:\s', r'([^\]]+)', r'\]'), (eps, keep), True, containswords=True),
                    complexmetafunction),
    CHAT_Annotation('Replacement of Real Word', '8.3:70', '10.3:73',
                    CHAT_ComplexRegex((r'\[::', r'([^\]]+)', r'\]'), (eps, keep), True), complexmetafunction),
    CHAT_Annotation('Alternative Transcription', '8.3:70', '10.3:74',
                    CHAT_ComplexRegex((r'\[=\?', r'([^\]]+)', r'\]'), (keep, eps), True), complexmetafunction),
    CHAT_Annotation('Dependent Tier on Main Line', '8.3:70', 'none',
                    CHAT_ComplexRegex((r'\[%\w\w\w:', anybutrb, r'\]'), (keep, eps), True), complexmetafunction),
    # @@must do something with the speaker
    CHAT_Annotation('Comment on Main Line', '8.3:70', '10.3:74',
                    CHAT_ComplexRegex((r'\[%\s+', anybutrb, r'\]'), (keep, eps), True), complexmetafunction),
    CHAT_Annotation('Best Guess', '8.3:70-71', '10.3:74', CHAT_SimpleScopedRegex(r'\[\?\]', keep, True, monadic),
                    simplescopedmetafunction),
    CHAT_Annotation('Repetition', '8.4:72', '10.4:75-76', CHAT_SimpleScopedRegex(r'\[/\]', eps, True, monadic),
                    simplescopedmetafunction),
    CHAT_Annotation('Multiple Repetition', '8.4:72-73', '10.4:76',
                    CHAT_ComplexRegex((r'\[x', r'[0-9]+', r'\]'), (keep, eps), True), complexmetafunction),
    CHAT_Annotation('Retracing', '8.4:73', '10.4:76-77', CHAT_SimpleScopedRegex(r'\[//\]', eps, True, monadic),
                    simplescopedmetafunction),
    CHAT_Annotation('Reformulation', '8.4:73-74', '10.4:77', CHAT_SimpleScopedRegex(r'\[///\]', eps, True, monadic),
                    simplescopedmetafunction),
    CHAT_Annotation('False Start Without Retracing', '8.4:74', '10.4:77',
                    CHAT_SimpleScopedRegex(r'\[/\-\]', eps, True, dyadic), simplescopedmetafunction),
    CHAT_Annotation('Unclear Retracing Type', '8.4:74', '10.4:77',
                    CHAT_SimpleScopedRegex(r'\[/\?\]', keep, True, monadic), simplescopedmetafunction),
    CHAT_Annotation('Excluded Material', '', '10.4:77-78', CHAT_SimpleScopedRegex(r'\[e\]', eps, True, monadic),
                    simplescopedmetafunction),
    CHAT_Annotation('Clause Delimiter', '8.4:74', '78', CHAT_SimpleRegex(r'\[\^c\]', eps, False),
                    simplemetafunction(identity)),  # needs extension
    CHAT_Annotation('Interposed Word', '8.4:74', '9.10.2:65', CHAT_SimpleRegex(r'&\*\w\w\w:[\w:]+', eps, False),
                    # grouped metadata would come in handy here ID100 text speaker = XXX, ID100 text interposedword = hmm
                    simplemetafunction(interposedword)),
    CHAT_Annotation('Postcode', '8.6:75', '10.5:78',
                    CHAT_ComplexRegex((r'\[\+\s+', wordpat, r'\]'), (keep, eps), False),
                    complexmetafunction),
    CHAT_Annotation('Language Precode', '8.6:75', '10.5:79',
                    CHAT_ComplexRegex((r'\[\-\s+', wordpat, r'\]'), (keep, eps), False),
                    complexmetafunction),
    CHAT_Annotation('Excluded Utterance', '8.6:75-76', '10.5:79', CHAT_SimpleRegex(r'\[\+\s+bch\]', eps, False),
                    simplemetafunction(interposedword)),
    CHAT_Annotation('Included Utterance', '8.6:76', '10.5:79', CHAT_SimpleRegex(r'\[\+\s+trn\]', eps, False),
                    simplemetafunction(interposedword)),
    CHAT_Annotation('Zero Utterance', '', '10.5:79, 11.1:81', CHAT_SimpleRegex(r'\b0\b', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('Segment Repetition', '10:85,11:89', '13:91', CHAT_InWordRegex(u'\u21AB.*?\u21AB', ''),
                    complexmetafunction),
    CHAT_Annotation('Joined Words', '6.6.4:46', '8.6.3:51', CHAT_InWordRegex(r'_', space), complexmetafunction),
    # take care extra token!@@
    CHAT_Annotation('Clitic Boundary', '6.6.15:52', 'not found', CHAT_InWordRegex(r'~', space), complexmetafunction),
    # take care extra token@@
    CHAT_Annotation('Blocked Segments', '10:85,11:89', '13:91', CHAT_InWordRegex(u'\u2260.*?\u2260', ''),
                    complexmetafunction),
    # these must be applied after [/], [//], [///] etc
    CHAT_Annotation('Untranscribed Material', '6.4:42', '8.4:47', CHAT_SimpleRegex(r'www', eps, False),
                    simplemetafunction(epsf)),
    CHAT_Annotation('Phonological Fragment', '6.4:42', '8.4:48',
                    CHAT_SimpleRegex(r'&' + simplewordpat, eps, False), simplemetafunction(identity)),
    CHAT_Annotation('Phonological Fragment', 'None', '8.4:48; https://talkbank.org/manuals/Clin-CLAN.pdf '
                                                     'states &+ for phonological fragments(p. 18)',
                    CHAT_SimpleRegex(r'&\+' + simplewordpat, eps, False), simplemetafunction(identity)),
    CHAT_Annotation('Filler', 'None', '8.4:48',
                    CHAT_SimpleRegex(r'&\-' + wordpat, eps, False), simplemetafunction(identity)),

    # ad-hoc extensiosn for Lotti
    CHAT_Annotation('[een]', 'ad-hoc extension', 'ad-hoc extension', CHAT_SimpleRegex(r'\[een\]', eps, False),
                    simplemetafunction(identity)),
    CHAT_Annotation('[twee]', 'ad-hoc extension', 'ad-hoc extension', CHAT_SimpleRegex(r'\[twee\]', eps, False),
                    simplemetafunction(identity)),

]


def findscope(tokenlist, offset=0):
    bregex = re.compile(r'^\<$')
    mregex = re.compile(r'^[^\<\>]+$')
    eregex = re.compile(r'^\>$')
    result1 = bracketseq(tokenlist, bregex, mregex, eregex)
    if result1 is None:
        result = None
    else:
        (b, e) = result1
        result = (b + offset, e + offset)
    return result


def bracketseq(tokenlist, bregex, mregex, eregex):
    state = bstate
    tokenctr = 0
    if tokenlist is None:
        return None
    for token in tokenlist:
        if state == bstate:
            if bregex.search(token.word):
                begin = tokenctr
                state = mstate
            else:
                pass
        elif state == mstate:
            if eregex.search(token.word) is not None:
                end = tokenctr
                state = estate
            elif bregex.search(token.word) is not None:
                settings.LOGGER.error('Range Open symbol encountered inside brackets in {}'.format(show(tokenlist)))
                state = mstate
            elif mregex.search(token.word) is not None:
                state = mstate
            else:
                settings.LOGGER.error('Incorrect element between brackets ({}) in: {}'.format(token.word, show(tokenlist)))
                state = mstate
        elif state == estate:
            break
        tokenctr += 1
    if state == estate:
        result = (begin, end)
    else:
        result = None
    return result


def anysearch(reset, word):
    result = False
    for regex in reset:
        if result:
            return result
        else:
            if regex.search(word):
                result = True
    return result


def isopenbracket(word):
    result = anysearch(openbrackets, word)
    return result


def isclosebracket(word):
    result = anysearch(closebrackets, word)
    return result


def get_CHATpatterns(annotations):
    result = set()
    # include scope begin and end
    result = result.union({r'\<', r'\>'})
    openbrackets = set()
    closebrackets = set()
    # include regular expressions from the annotations
    for annotation in annotations:
        newpats = set()
        theregex = annotation.regex
        if isinstance(theregex, CHAT_SimpleRegex):
            newpats = {theregex.regex}
        elif isinstance(theregex, CHAT_SimpleScopedRegex):
            newpats = {theregex.regex}
        elif isinstance(theregex, CHAT_ComplexRegex):
            newpats = {theregex.regexbegin, theregex.regexend}
            openbrackets = openbrackets.union({theregex.compiledrebegin})
            closebrackets = closebrackets.union({theregex.compiledreend})
        elif isinstance(theregex, CHAT_InWordRegex):
            newpats = {}
        else:
            settings.LOGGER.error('Unknown Regex type: {}'.format(theregex))
            newpats = {}
        result = result.union(newpats)
    return (result, openbrackets, closebrackets)


(CHAT_patterns, openbrackets, closebrackets) = get_CHATpatterns(annotations)



'''
The generatemacros module provides the function generatemacros of type None -> Dict[str, str], which generates a macrodictionary
for systematically constructed macro definitions.

In particular, the module generatemacros.py generates macros for a list of pairs (verb, adposition),
e.g. (*slaan*, *op*) in which Alpino analyzes the adposition as the head of an prepositional complement (pc),
but where it should be considered the head of a modifier. These cases should also count as such if the
adposition is written together with *er*, *hier* or *daar* (e.g. *erop*, *hierop*, *daarop*),
which are considered unanalysable adverbs by Alpino.
So we need a macro which characterizes these exceptions for each (verb, adposition) pair and, for each pair,
for the combinations of the adposition with *er*, *hier* en *daar* when written together with the adposition.

The generatemacros function generates this macro on the basis of a list of  (verb, adposition) pairs contained in this module.
[These should probably better be moved to an exceptionslexicon module.]

Running the module standalone outputs the macrodefinitions to stdout
'''

from typing import Dict, List, Set, Tuple

StrPair = Tuple[str, str]

tarsp_wvzexceptions: Set[StrPair] = {('hebben', 'van'),  # hebben can occur with pc/van but only in fixed expressions: last hebben van, spijt hebben van
                                     ('slaan', 'op'),  # of course correct but not used in this sense with very young children
                                     ('lukken', 'in'), \
                                     ('passen', 'bij'), \
                                     ('vallen', 'vanaf'),
                                     # next list derived from CHILDES via Paqu
                                     # query: https://paqu.let.rug.nl:8068/?db=childesdutch&word=&rel=pc&hword=&postag=vz&hpostag=ww&meta=&sn=10
                                     # and manual filtering see file Tarsp_wrongpcs.txt in the lassy folder
                                     ('af_brengen', 'vanaf'),
                                     ('af_komen', 'op'),
                                     ('af_komen', 'van'),
                                     ('af_komen', 'vanaf'),
                                     ('bouwen', 'op'),
                                     ('dansen', 'op'),
                                     ('doen', 'overheen'),
                                     ('door_gaan', 'op'),
                                     ('draaien', 'om'),
                                     ('eten', 'van'),
                                     ('gaan', 'aan'),
                                     ('gaan', 'om'),
                                     ('gaan', 'overheen'),
                                     ('gaan', 'tussendoor'),
                                     ('gaan', 'voor'),
                                     ('hebben', 'af'),
                                     ('hebben', 'uit'),
                                     ('helpen', 'bij'),
                                     ('houden', 'voor'),
                                     ('komen', 'achteraan'),
                                     ('komen', 'af'),
                                     ('komen', 'bij'),
                                     ('komen', 'doorheen'),
                                     ('komen', 'om'),
                                     ('komen', 'tot'),
                                     ('krijgen', 'af'),
                                     ('krijgen', 'op'),
                                     ('kunnen', 'omheen'),
                                     ('liggen', 'bij'),
                                     ('mee_gaan', 'met'),
                                     ('plakken', 'aan'),
                                     ('rijden', 'op'),
                                     ('sabbelen', 'op'),
                                     ('spelen', 'in'),
                                     ('staan', 'op'),
                                     ('staan', 'voor'),
                                     ('struikelen', 'over'),
                                     ('tellen', 'tot'),
                                     ('uit_komen', 'in'),
                                     ('varen', 'op'),
                                     ('vast_knopen', 'aan'),
                                     ('zeggen', 'in '),
                                     ('zien', 'op'),
                                     ('zijn', 'tegen'),
                                     ('zijn', 'voor'),
                                     ('zitten', 'op'),
                                     ('zitten', 'voor'),
                                     ('zoeken', 'op')
                                     }

# also do erin erbij etc

Rlemma: str = """( @lemma="er{vz}" or @lemma="daar{vz}" or @lemma="hier{vz}" ) """

Tarsp_WVz_exception_basemodel: str = """(@pt="ww"  and @rel ="hd" and @lemma="{ww}" and
     ../node[@rel="pc" and
            ( (node[@rel="hd" and @lemma="{vz}"]) or
                ( @lemma="er{rvz}" or @lemma="daar{rvz}" or @lemma="hier{rvz}" )
            )
            ]
 )"""

pc_vc_exception_basemodel: str = """( @rel="pc" and node[@rel = "hd" and @lemma="{vz}"] and ../node[@rel="hd" and @pt="ww" and @lemma="{ww}"] )"""

macrodef_model: str = '{name} = """{exp}"""\n'


def generatemacros() -> Dict[str, str]:
    newmacros: Dict[str, str] = {}
    newparts: List[str] = []
    for (ww, vz) in tarsp_wvzexceptions:
        if vz == 'tot':
            rvz = 'toe'
            newpart = Tarsp_WVz_exception_basemodel.format(ww=ww, vz=rvz, rvz=rvz)  # voor : hij werkte er niet mee
            newparts.append(newpart)
        elif vz == 'met':
            rvz = 'mee'
            newpart = Tarsp_WVz_exception_basemodel.format(ww=ww, vz=rvz, rvz=rvz)  # voor : hij kwam er niet toe
            newparts.append(newpart)
        else:
            rvz = vz
        newpart = Tarsp_WVz_exception_basemodel.format(ww=ww, vz=vz, rvz=rvz)
        newparts.append(newpart)
    partsor = ' or '.join(newparts)
    newmacro = "( {} )".format(partsor)
    newmacros['Tarsp_WVz_exception'] = newmacro

    newparts = []
    for (ww, vz) in tarsp_wvzexceptions:
        newpart = pc_vc_exception_basemodel.format(ww=ww, vz=vz)
        newparts.append(newpart)
    partsor = ' or '.join(newparts)
    newmacro = "( {} )".format(partsor)
    newmacros['Tarsp_pc_vc_exception'] = newmacro
    return newmacros


def main():
    macros = generatemacros()
    for macroname in macros:
        macrodef = macrodef_model.format(name=macroname, exp=macros[macroname])
        print(macrodef)


if __name__ == '__main__':
    main()



'''
The script *checkcorrections* compares the errorlogging file with the error reference file

The error reference file can be found in the bronze subfolder of a dataset and is called <dataset>_error_ref.xlsx.

Warning: The error reference files are attempts to get to reference files. They mainly check for precision,
not for recall. They will very likely require several new versions.
'''

import os
from optparse import OptionParser

from sastadev.conf import settings
from sastadev.constants import bronzefolder, loggingfolder, resultsfolder
from sastadev.stringfunctions import nono
from sastadev.xlsx import getxlsxdata, mkworkbook


def main():

    parser = OptionParser()
    parser.add_option("-d", "--dataset", dest="dataset",
                      help="Dataset to be dealt with")
    parser.add_option("-r", "--reference", dest="referencepath",
                      help="Path to the referencedata")

    (options, args) = parser.parse_args()

    if nono(options.dataset):
        print('No dataset specified. Aborting')
        exit(-1)

    if nono(options.referencepath):

        options.referencepath = os.path.join(settings.DATAROOT, options.dataset, bronzefolder)

    missedcorrectionslist = []
    originalsents = {}

    resultspath = os.path.join(settings.DATAROOT, options.dataset, resultsfolder)
    loggingpath = os.path.join(settings.DATAROOT, options.dataset, loggingfolder)

    dataprefix = options.dataset

    errorloggingfilename = dataprefix + '_errorlogging.xlsx'
    errorloggingfullname = os.path.join(loggingpath, errorloggingfilename)

    errorreffilename = dataprefix + '_error_ref.xlsx'
    errorreffullname = os.path.join(options.referencepath, errorreffilename)

    logheader, logdata = getxlsxdata(errorloggingfullname)
    refheader, refdata = getxlsxdata(errorreffullname)

    refdict = {(row[0].lower(), row[1]): row[3] for row in refdata}

    correctcorrections = 0
    missedcorrections = 0
    wrongcorrections = 0
    for row in logdata:
        key = (row[0].lower(), row[5])
        if 'Original' in row[4]:
            originalsents[key] = row[7]
        if 'BEST' in row[10]:
            logsent = row[9]
            if key not in refdict:
                print('Missing example in refdict: {}'.format(key))
                print(row[9])
                missedcorrections += 1
                origsent = originalsents[key] if key in originalsents else ''
                newmissedcorrection = [row[0], row[5], origsent, logsent]
                missedcorrectionslist.append(newmissedcorrection)
            else:
                refsent = refdict[key]
                if refsent != logsent:
                    print('Mismatch: {}'.format(key))
                    print('refsent=<{}>'.format(refsent))
                    print('logsent=<{}>'.format(logsent))
                    wrongcorrections += 1
                else:
                    correctcorrections += 1

    allcorrections = correctcorrections + wrongcorrections + missedcorrections

    correctioncounts = [correctcorrections, wrongcorrections, missedcorrections]
    labels = ['correct corrections', 'wrong corrections', 'missed corrections']
    labeled_corrections = zip(labels, correctioncounts)

    print('\nSummary:\n')
    for label, corr in labeled_corrections:
        print('{} = {} ({:.2f}%)'.format(label, corr, corr / allcorrections * 100))

    refaddfilename = f'{options.dataset}_newrefcandidates.xlsx'
    refadditions = os.path.join(options.referencepath, refaddfilename)
    wb = mkworkbook(refadditions, [refheader], missedcorrectionslist, freeze_panes=(0, 1))
    wb.close()


if __name__ == '__main__':
    main()
